"""Module containing routines for the validation framework"""

from __future__ import annotations

from collections import defaultdict
from dataclasses import fields
from typing import Any, Callable, Dict, List, Optional, Tuple

import pandas as pd
import swifter  # noqa #pylint: disable=unused-import
from tqdm import tqdm

from rxnutils.chem.utils import join_smiles_from_reaction, split_rsmi, split_smiles_from_reaction

ActionType = Callable[[pd.DataFrame], pd.DataFrame]

_REGISTERED_ACTIONS: Dict[str, ActionType] = {}


def action(obj: ActionType) -> ActionType:
    """
    Decorator that register a callable as a validation action.

    An action will be called with a `pandas.DataFrame` object
    and return a new `pandas.DataFrame` object.

    An action needs to have an attribute `pretty_name`.

    :param obj: the callable to register as an action
    :return: the same as `obj`.
    """
    global _REGISTERED_ACTIONS
    if not hasattr(obj, "pretty_name"):
        raise ValueError("A validation action needs to have an attribue 'pretty_name'")
    if not callable(obj):
        raise ValueError("A validation action needs to be callable")
    _REGISTERED_ACTIONS[obj.pretty_name] = obj
    return obj


def list_actions(short: bool = False) -> None:
    """List all available actions in a nice table"""
    dict_ = defaultdict(list)
    for name, obj in _REGISTERED_ACTIONS.items():
        dict_["name"].append(name)
        optionals = []
        requireds = []
        kwargs = {}
        for field in fields(obj):
            if "MISSING" in repr(field.default):
                requireds.append(f"{field.name} ({field.type})")
                kwargs[field.name] = ""
            else:
                optionals.append(f"{field.name} ({field.type})")
        if not short:
            dict_["required arguments"].append(", ".join(requireds))
            dict_["optional arguments"].append(", ".join(optionals))
        dict_["description"].append(str(obj(**kwargs)))
    print(pd.DataFrame(dict_).to_string(index=False, max_colwidth=100))


def create_action(pretty_name: str, *args: Any, **kwargs: Any) -> ActionType:
    """
    Create an action that can be called

    :param pretty_name: the name of the action
    :return: the instantiated actions
    """
    global _REGISTERED_ACTIONS
    if pretty_name not in _REGISTERED_ACTIONS:
        # Check for pretty_name variance...
        # We can have the same action more than once, hopefuly with different parameters...
        for action_pretty_name, reg_action in _REGISTERED_ACTIONS.items():
            if pretty_name.startswith(action_pretty_name):
                return reg_action(*args, **kwargs)
        raise KeyError(f"Unknown action: {pretty_name}")
    return _REGISTERED_ACTIONS[pretty_name](*args, **kwargs)


class _DataFrameApplier:
    def __init__(self) -> None:
        self.max_workers: Optional[int] = None
        tqdm.pandas()

    def __call__(self, data: pd.DataFrame, *args: Any, **kwargs: Any) -> Any:
        if self.max_workers is None:
            return data.swifter.apply(*args, **kwargs)
        if self.max_workers == 1:
            return data.apply(*args, **kwargs)
        return data.swifter.set_npartitions(self.max_workers).apply(*args, **kwargs)


global_apply = _DataFrameApplier()


class ReactionActionMixIn:
    """Mixin class with standard routines for splitting and joining reaction SMILES"""

    def join_lists(
        self,
        reactants_list: List[str],
        reagents_list: List[str],
        products_list: List[str],
    ) -> str:
        """
        Join list of components into a reaction SMILES

        :param reactants_list: the list of reactant SMILES
        :param reagents_list: the list of reagent SMILES
        :param products_list: the list of product SMILES
        :return: the concatenated reaction SMILES
        """
        return self.join_smiles(
            join_smiles_from_reaction(reactants_list),
            join_smiles_from_reaction(reagents_list),
            join_smiles_from_reaction(products_list),
        )

    def join_smiles(self, reactants: str, reagents: str, products: str) -> str:
        """
        Join component SMILES into a reaction SMILES

        :param reactants: the reactant SMILES
        :param reagents: the reagent SMILES
        :param products: the product SMILES
        :return: the concatenated reaction SMILES
        """
        return ">".join([reactants, reagents, products])

    def split_lists(self, row: pd.Series) -> Tuple[List[str], List[str], List[str]]:
        """
        Split a reaction SMILES into list of component SMILES

        :param row: the row with the SMILES
        :return: the list of SMILES of the components
        """
        reactants, reagents, products = self.split_smiles(row)

        reactants_list = split_smiles_from_reaction(reactants)
        reagents_list = split_smiles_from_reaction(reagents)
        products_list = split_smiles_from_reaction(products)
        return reactants_list, reagents_list, products_list

    def split_smiles(self, row: pd.Series) -> Tuple[str, str, str]:
        """
        Split a reaction SMILES into components SMILES

        :param row: the row with the SMILES
        :return: the SMILES of the components
        """
        return split_rsmi(row[self.in_column])



""" Module contain method to compute distance matrix using TED """

import time
from typing import Sequence

import numpy as np

from rxnutils.routes.base import SynthesisRoute
from rxnutils.routes.ted.reactiontree import ReactionTreeWrapper


def ted_distances_calculator(
    routes: Sequence[SynthesisRoute], content: str = "both", timeout: int = None
) -> np.ndarray:
    """
    Compute the TED distances between each pair of routes

    :param routes: the routes to calculate pairwise distance on
    :param content: determine what part of the synthesis trees to include in the calculation,
            options 'molecules', 'reactions' and 'both', default 'both'
    :param timeout: if given, raises an exception if timeout is taking longer time
    :return: the square distance matrix
    """
    distances = np.zeros([len(routes), len(routes)])
    distance_wrappers = [ReactionTreeWrapper(route, content) for route in routes]
    time0 = time.perf_counter()
    for i, iwrapper in enumerate(distance_wrappers):
        # fmt: off
        for j, jwrapper in enumerate(distance_wrappers[i + 1:], i + 1):
            distances[i, j] = iwrapper.distance_to(jwrapper)
            distances[j, i] = distances[i, j]
        # fmt: on
        time_past = time.perf_counter() - time0
        if timeout is not None and time_past > timeout:
            raise ValueError(f"Unable to compute distance matrix in {timeout} s")
    return distances



""" Routines for scoring synthesis routes
"""

from typing import Any, Callable, Dict, List, Tuple

import numpy as np

from rxnutils.routes.base import SynthesisRoute
from rxnutils.routes.deepset.scoring import DeepsetModelClient  # noqa
from rxnutils.routes.deepset.scoring import deepset_route_score
from rxnutils.routes.retro_bleu.scoring import ngram_overlap_score  # noqa
from rxnutils.routes.retro_bleu.scoring import retro_bleu_score


def route_sorter(
    routes: List[SynthesisRoute], scorer: Callable[..., float], **kwargs: Any
) -> Tuple[List[SynthesisRoute], List[float]]:
    """
    Scores and sort a list of routes.
    Returns a tuple of the sorted routes and their scores.

    :param routes: the routes to score
    :param scorer: the scorer function
    :param kwargs: additional argument given to the scorer
    :return: the sorted routes and their scores
    """
    scores = np.asarray([scorer(route, **kwargs) for route in routes])
    sorted_idx = np.argsort(scores)
    routes = [routes[idx] for idx in sorted_idx]
    return routes, scores[sorted_idx].tolist()


def route_ranks(scores: List[float]) -> List[int]:
    """
    Compute the rank of route scores. Rank starts at 1

    :param scores: the route scores
    :return: a list of ranks for each route
    """
    ranks = [1]
    for idx in range(1, len(scores)):
        if abs(scores[idx] - scores[idx - 1]) < 1e-8:
            ranks.append(ranks[idx - 1])
        else:
            ranks.append(ranks[idx - 1] + 1)
    return ranks


def badowski_route_score(
    route: SynthesisRoute,
    mol_costs: Dict[bool, float] = None,
    average_yield: float = 0.8,
    reaction_cost: float = 1.0,
) -> float:
    """
    Calculate the score of route using the method from
    (Badowski et al. Chem Sci. 2019, 10, 4640).

    The reaction cost is constant and the yield is an average yield.
    The starting materials are assigned a cost based on whether they are in
    stock or not. By default starting material in stock is assigned a
    cost of 1 and starting material not in stock is assigned a cost of 10.

    To be accurate, each molecule node need to have an extra
    boolean property called `in_stock`.

    :param route: the route to analyze
    :param mol_costs: the starting material cost
    :param average_yield: the average yield, defaults to 0.8
    :param reaction_cost: the reaction cost, defaults to 1.0
    :return: the computed cost
    """

    def traverse(tree_dict):
        mol_cost = mol_costs or {True: 1, False: 10}

        reactions = tree_dict.get("children", [])
        if not reactions:
            return mol_cost[tree_dict.get("in_stock", True)]

        child_sum = sum(1 / average_yield * traverse(child) for child in reactions[0]["children"])
        return reaction_cost + child_sum

    return traverse(route.reaction_tree)


def reaction_class_rank_score(
    route: SynthesisRoute,
    reaction_class_ranks: Dict[str, int],
    preferred_classes: List[str],
    non_preferred_factor: float = 0.25,
) -> float:
    """
    Calculates a score of a route based on the reaction class rank score, i.e.
    how likely a particular reaction class is to succeed.

    Each step in the route is scored based on the following factors:
        * The reaction class rank
        * The step in the synthesis sequence
        * The preference of the reaction class

    The score is min-max normalized relative to the maximum depth of the three
    and the max/min of the class ranks.

    :param route: the route to score
    :param reaction_class_ranks: the rank score of NextMove classes
    :param preferred_classes: the preferred reaction classes
    :param non_preferred_factor: steps with non-preferred classes are multiplied by this
    :return: the computed score
    """

    def _score_reaction_tree(tree_dict):
        cls = tree_dict["metadata"]["classification"]
        if " " in cls:
            cls = cls.split(" ")[0]
        if _nextmove_superclass(cls) in preferred_superclasses:
            pref_factor = 1.0
        else:
            pref_factor = non_preferred_factor
        step_weight = tree_dict["metadata"]["forward_step"]
        score = pref_factor * step_weight * reaction_class_ranks.get(cls, min_rank)

        child_scores = []
        for child in tree_dict["children"]:
            grandchildren = child.get("children", [])
            if grandchildren:
                child_score = _score_reaction_tree(grandchildren[0])
                child_scores.append(child_score)
        if child_scores:
            score += min(child_scores)

        return score

    min_rank = min(reaction_class_ranks.values())
    max_step_weight = sum(range(1, route.max_depth + 1))
    min_score = max_step_weight * min_rank * non_preferred_factor
    max_score = max_step_weight * max(reaction_class_ranks.values())

    preferred_superclasses = [_nextmove_superclass(cls) for cls in preferred_classes]
    first_reaction = route.reaction_tree["children"][0]
    score = _score_reaction_tree(first_reaction)
    return (score - min_score) / (max_score - min_score)


def _nextmove_superclass(cls: str) -> str:
    """Extracting the NextMove superclass"""
    cls = str(cls).strip()
    if cls == "nan":
        return cls
    if cls.count(".") == 1:
        return cls
    return ".".join(cls.split(".")[:2])



"""Module containing various chemical utility routines"""

import functools
import logging
import re
from collections import defaultdict
from typing import List, Optional, Tuple

import numpy as np
import rdchiral.template_extractor
from rdkit import Chem
from rdkit.Chem import AllChem, Descriptors
from rdkit.Chem.MolStandardize import rdMolStandardize
from rdkit.Chem.SaltRemover import SaltRemover

# Pattern for spliting Reaction SMILES should be '>' not prefixed by '-' (i.e. '->')
rsmi_split_pattern = re.compile(r"(?<!-)>")


def get_symmetric_sites(mol: Chem.rdchem.Mol, candidate_atoms: List[int]) -> List[List[int]]:
    """
    Get all symmetric sites (atoms) of each atom in the list of atoms defined by their
    atomic index (not atom-map number). Symmetry is assessed with respect to the molecule.
    Symmetric atoms will have different atom inds but the same rank index from
    CanonicalRankAtoms.

    :param mol: RdKit molecule
    :param candidate_atoms: Indices of the atoms that will be checked for symmetry.
    :return: A list of all symmetric sites (list of atom-ids) that include the candidate
        atoms. Returns empty list if no atoms have symmetric sites.
    """
    n_atoms = mol.GetNumAtoms()
    if any(atom_idx >= n_atoms for atom_idx in candidate_atoms):
        raise ValueError(
            f"At least one candidate atom-idx is out of bounds in molecule with {n_atoms} atoms: {candidate_atoms}"
        )

    sites = defaultdict(lambda: [])
    for atom_idx, rank_idx in enumerate(list(Chem.CanonicalRankAtoms(mol, breakTies=False))):
        sites[rank_idx].append(atom_idx)

    symmetric_sites = [
        atoms for atoms in sites.values() if len(atoms) > 1 and any(atom_idx in candidate_atoms for atom_idx in atoms)
    ]
    return symmetric_sites


def get_mol_weight(smiles: str) -> Optional[float]:
    """Calculate molecule's exact molecular weight.
    :param smiles: Molecule's SMILES.
    :return: Molecule's exact molecular weight.
    """
    mol = Chem.MolFromSmiles(smiles)
    if not mol:
        return None
    mol_weight = Descriptors.ExactMolWt(mol)

    return np.round(mol_weight, 6)


def get_special_groups(mol) -> List[Tuple[Tuple[int, ...], Tuple[int, ...]]]:
    """
    Given an RDKit molecule, this function returns a list of tuples, where
    each tuple contains the AtomIdx's for a special group of atoms which should
    be included in a fragment all together. This should only be done for the
    reactants, otherwise the products might end up with mapping mismatches
    We draw a distinction between atoms in groups that trigger that whole
    group to be included, and "unimportant" atoms in the groups that will not
    be included if another atom matches.
    """
    # Define templates
    group_templates = [
        (
            range(3),
            "[OH0,SH0]=C[O,Cl,I,Br,F]",
        ),  # carboxylic acid / halogen
        (
            range(3),
            "[OH0,SH0]=CN",
        ),  # amide/sulfamide
        (
            range(4),
            "S(O)(O)[Cl]",
        ),  # sulfonyl chloride
        (
            range(3),
            "B(O)O",
        ),  # boronic acid/ester
        ((0,), "[Si](C)(C)C"),  # trialkyl silane
        ((0,), "[Si](OC)(OC)(OC)"),  # trialkoxy silane, default to methyl
        (
            range(3),
            "[N;H0;$(N-[#6]);D2]-,=[N;D2]-,=[N;D1]",
        ),  # azide
        (
            range(8),
            "O=C1N([Br,I,F,Cl])C(=O)CC1",
        ),  # NBS brominating agent
        (range(11), "Cc1ccc(S(=O)(=O)O)cc1"),  # Tosyl
        ((7,), "CC(C)(C)OC(=O)[N]"),  # N(boc)
        ((4,), "[CH3][CH0]([CH3])([CH3])O"),  # t-Butyl
        (
            range(2),
            "[C,N]=[C,N]",
        ),  # alkene/imine
        (
            range(2),
            "[C,N]#[C,N]",
        ),  # alkyne/nitrile
        (
            (2,),
            "C=C-[*]",
        ),  # adj to alkene
        (
            (2,),
            "C#C-[*]",
        ),  # adj to alkyne
        (
            (2,),
            "O=C-[*]",
        ),  # adj to carbonyl
        ((3,), "O=C([CH3])-[*]"),  # adj to methyl ketone
        (
            (3,),
            "O=C([O,N])-[*]",
        ),  # adj to carboxylic acid/amide/ester
        (
            range(4),
            "ClS(Cl)=O",
        ),  # thionyl chloride
        (
            range(2),
            "[Mg,Li,Zn,Sn][Br,Cl,I,F]",
        ),  # grinard/metal (non-disassociated)
        (
            range(3),
            "S(O)(O)",
        ),  # SO2 group
        (
            range(2),
            "N~N",
        ),  # diazo
        (
            (1,),
            "[!#6;R]@[#6;R]",
        ),  # adjacency to heteroatom in ring
        (
            (2,),
            "[a!c]:a:a",
        ),  # two-steps away from heteroatom in aromatic ring
        # ((1,), 'c(-,=[*]):c([Cl,I,Br,F])',), # ortho to halogen on ring - too specific?
        # ((1,), 'c(-,=[*]):c:c([Cl,I,Br,F])',), # meta to halogen on ring - too specific?
        ((0,), "[B,C](F)(F)F"),  # CF3, BF3 should have the F3 included
    ]

    # Stereo-specific ones (where we will need to include neighbors)
    # Tetrahedral centers should already be okay...
    group_templates += [
        (
            (
                1,
                2,
            ),
            "[*]/[CH]=[CH]/[*]",
        ),  # trans with two hydrogens
        (
            (
                1,
                2,
            ),
            "[*]/[CH]=[CH]\\[*]",
        ),  # cis with two hydrogens
        (
            (
                1,
                2,
            ),
            "[*]/[CH]=[CH0]([*])\\[*]",
        ),  # trans with one hydrogens
        (
            (
                1,
                2,
            ),
            "[*]/[D3;H1]=[!D1]",
        ),  # specified on one end, can be N or C
    ]

    # Amol addition
    # Carried over from previous iteration (before merge with rdchiral)
    # There is some redundancy from merge with rdchiral and templates are not general
    group_templates += [
        (range(3), "C(=O)Cl"),  # acid chloride
        (range(4), "[$(N-!@[#6])](=!@C=!@O)"),  # isocyanate
        (range(2), "C=O"),  # carbonyl
        (range(4), "ClS(Cl)=O"),  # thionyl chloride
        (range(5), "[#6]S(=O)(=O)[O]"),  # RSO3 leaving group
        (range(5), "[O]S(=O)(=O)[O]"),  # SO4 group
        (range(3), "[N-]=[N+]=[C]"),  # diazo-alkyl
        (
            range(15),
            "[#6][C]([#6])([#6])[#8]-[#6](=[O])-[#8]-[#6](=O)-[#8]C([#6])([#6])[#6]",
        ),  # Boc Anhydride
        # Protecting Groups
        # Amino
        (range(18), "[#7]-[#6](=O)-[#8]-[#6]-[#6]-1-c2ccccc2-c2ccccc-12"),  # Fmoc
        (range(8), "[#6]C([#6])([#6])[#8]-[#6](-[#7])=O"),  # Boc
        (range(11), "[#7]-[#6](=O)-[#8]-[#6]-c1ccccc1"),  # Cbz
        (range(6), "[#7]-[#6](=O)-[#8]-[#6]=[#6]"),  # Voc
        (range(7), "[#7]-[#6](=O)-[#8]-[#6]-[#6]=[#6]"),  # Alloc
        (range(4), "[#6]-[#6](-[#7])=O"),  # Acetamide
        (range(7), "[#7]-[#6](=O)C(F)(F)F"),  # Trifluoroacetamide
        (range(7), "[#7]-[#6](=O)C(Cl)(Cl)Cl"),  # Trichloroacetamide
        (range(11), "O=[#6]-1-[#7]-[#6](=O)-c2ccccc-12"),  # Phthalimide
        (range(8), "[#7]-[#6]-c1ccccc1"),  # Benzylamine
        (range(10), "[#6]-[#8]-c1ccc(-[#6]-[#7])cc1"),  # p-Methoxybenzyl (PMB)
        (range(9), "[#6]-[#8]-c1ccc(-[#7])cc1"),  # PMP
        (range(20), "[#7]C(c1ccccc1)(c1ccccc1)c1ccccc1"),  # Triphenylmethylamine
        (range(8), "[#7]=[#6]-c1ccccc1"),  # N-Benzylidene
        (range(11), "[#6]-c1ccc(cc1)S([#7])(=O)=O"),  # p-Toluenesulfonamide
        (range(13), "[#7]S(=O)(=O)c1ccc(cc1)-[#7+](-[#8-])=O"),  # p-Nosylsulfonamide
        (range(5), "[#6]S([#7])(=O)=O"),  # Mesylate
        (
            range(7),
            "[#7]-[#6]-1-[#6]-[#6]-[#6]-[#6]-[#8]-1",
        ),  # N-Tetrahydropyranyl ether (N-THP)
        (range(14), "[#7]=[#6](-c1ccccc1)-c1ccccc1"),  # N-Benzhydrylidene
        (range(14), "[#7]-[#6](-c1ccccc1)-c1ccccc1"),  # N-Benzhydryl
        (range(5), "[#6]-[#7](-[#6])-[#6]=[#7]"),  # Dimethylaminomethylene
        (range(9), "[#7]-[#6](=O)-c1ccccc1"),  # N-Benzoate (N-Bz)
        (range(8), "[#7]=[#6]-c1ccccc1"),  # N-Benzylidene
        # Carbonyl
        (range(6), "[#6]-[#8]-[#6](-[#6])-[#8]-[#6]"),  # Dimethyl acetal
        (range(6), "[#6]-1-[#6]-[#8]-[#6]-[#8]-[#6]-1"),  # 1,3-Dioxane
        (range(5), "[#6]-1-[#6]-[#8]-[#6]-[#8]-1"),  # 1,3-Dioxolane
        (range(5), "[#6]-1-[#6]-[#16]-[#6]-[#16]-1"),  # 1,3-Dithiane
        (range(6), "[#6]-1-[#6]-[#16]-[#6]-[#16]-[#6]-1"),  # 1,3-Dithiolane
        (range(5), "[#6]-[#7](-[#6])-[#7]=[#6]"),  # N,N-Dimethylhydrazone
        (range(8), "[#6]C1([#6])[#6]-[#8]-[#6]-[#8]-[#6]1"),  # 1,3-Dioxolane (dimethyl)
        # Carboxyl
        (range(5), "[#6]-[#8]-[#6](-[#6])=O"),  # Methyl ester
        (range(6), "[#6]-[#6]-[#8]-[#6](-[#6])=O"),  # Ethyl ester
        (range(8), "[#6]-[#6](=O)-[#8]C([#6])([#6])[#6]"),  # t-Butyl ester
        (
            range(11),
            "[#6]-[#6](=O)-[#8]-[#6]-[#6]-1=[#6]-[#6]=[#6]-[#6]=[#6]-1",
        ),  # Benzyl ester
        (range(8), "[#6]-[#6](=O)-[#8][Si]([#6])([#6])[#6]"),  # Silyl Ester
        (range(8), "[#6]-[#6](=O)-[#16]C([#6])([#6])[#6]"),  # S-t-Butyl ester
        (range(6), "[#6]-[#6]-1=[#7]-[#6]-[#6]-[#8]-1"),  # 2-Alkyl-1,3-oxazoline
        # Hydroxyl
        (range(5), "[#6]-[#8]-[#6]-[#8]-[#6]"),  # Methoxymethyl ether
        (
            range(8),
            "[#6]-[#8]-[#6]-[#6]-[#8]-[#6]-[#8]-[#6]",
        ),  # 2-Methoxyethoxymethyl ether
        (
            range(9),
            "[#8]-[#6](=O)-[#8]-[#6]C(Cl)(Cl)Cl",
        ),  # 2,2,2-Trichloroethyl carbonate
        (range(7), "[#6]-[#8]C([#6])([#6])[#8]-[#6]"),  # Methoxypropyl acetal
        (range(7), "[#6]-[#6]-[#8]-[#6](-[#6])-[#8]-[#6]"),  # Ethoxyethyl acetal
        (
            range(11),
            "[#6]-[#8]-[#6]-[#8]-[#6]-[#6]-1=[#6]-[#6]=[#6]-[#6]=[#6]-1",
        ),  # Benzyloxymethyl acetal
        (
            range(8),
            "[#6]-[#8]-[#6]-1-[#6]-[#6]-[#6]-[#6]-[#8]-1",
        ),  # Tetrahydropyranyl ether
        (range(10), "[#6]-[#8]-[#6](=O)-c1ccccc1"),  # Benzoate (O-Bz)
        (range(6), "[#6]-[#8]C([#6])([#6])[#6]"),  # t-Butyl ether
        (range(5), "[#6]-[#8]-[#6]-[#6]=[#6]"),  # Allyl ether
        (range(13), "[#6]-[#8]-[#6]-c1ccc2ccccc2c1"),  # Napthyl ether
        (range(11), "[#6]-[#8]-[#6]-c1ccc(-[#8]-[#6])cc1"),  # p-Methoxybenzyl ether
        (range(6), "[#6]-[#8][Si]([#6])([#6])[#6]"),  # Trimethylsilyl ether
        (
            range(8),
            "[#6]-[#6](-[#6])-[#6][Si]([#6])([#6])[#8]",
        ),  # dimethylisopropylsilyl ether
        (
            range(9),
            "[#6]-[#6][Si]([#6]-[#6])([#6]-[#6])[#8]-[#6]",
        ),  # Triethylsilyl ether
        (
            range(12),
            "[#6]-[#8][Si]([#6](-[#6])-[#6])([#6](-[#6])-[#6])[#6](-[#6])-[#6]",
        ),  # Triisopropylsilyl ether
        (
            range(9),
            "[#6]-[#8][Si]([#6])([#6])C([#6])([#6])[#6]",
        ),  # t-Butyldimethylsilyl ether
        (
            range(10),
            "[#6]-[#8]-[#6]-[#8]-[#6]-[#6][Si]([#6])([#6])[#6]",
        ),  # [2-(trimethylsilyl)ethoxy]methyl
        (
            range(19),
            "[#6]-[#8][Si]([#6]-1=[#6]-[#6]=[#6]-[#6]=[#6]-1)([#6]-1=[#6]-[#6]=[#6]-[#6]=[#6]-1)C([#6])([#6])[#6]",
        ),  # t-Butyldiphenylsilyl ether
        (
            range(9),
            "[#6]-[#8][Si]([#6])([#6])C([#6])([#6])[#6]",
        ),  # t-Butyldimethylsilyl ether
        (
            range(20),
            "[#6]-[#6]-[#8][Si]([#6]-[#6]-[#6]-c1ccc(-[#6])cc1)([#6](-[#6])-[#6])[#6](-[#6])-[#6]",
        ),  # propyl-p-toluenediisopropylsilyl ether
        (range(5), "[#6]-[#8]-[#6](-[#6])=O"),  # Acetic acid ester (acetate)
        (range(8), "[#6]-[#8]-[#6](=O)C([#6])([#6])[#6]"),  # Pivalic acid ester
        (range(10), "[#6]-[#8]-[#6](=O)-c1ccccc1"),  # Benzoic acid ester
        (range(9), "[#6]-[#8]-[#6]-c1ccccc1"),  # Benzyl ether (O-Bn)
        # 1',2- 1',3-Diols
        (range(8), "[#6]C1([#6])[#8]-[#6]-[#6]-[#6]-[#8]1"),  # Acetonide
        (
            range(12),
            "[#6]-1-[#6]-[#8]-[#6](-[#8]-[#6]-1)-c1ccccc1",
        ),  # Benzylidene acetal
        # Alkyne
        (range(6), "[#6][Si]([#6])([#6])C#C"),  # Trimethylsilyl ether (TMS-alkyne)
        (
            range(9),
            "[#6]-[#6][Si]([#6]-[#6])([#6]-[#6])C#C",
        ),  # Triethylsilyl ether (TES-alkyne)
        # Thiol
        (range(5), "[#6]-[#16]-[#6](-[#6])=O"),  # S-Carbonyl deprotection (thioester)
        (
            range(15),
            "[#6]-[#6]-[#6](-[#6]-[#6])-[#6]C1([#6]-[#6]-[#6]-[#6]-[#6]1)[#6](-[#16])=O",
        ),  # S-Carbonyl deprotection (thioester)
        (
            range(9),
            "[#6]-[#6]-[#7](-[#6]-[#6])-[#6](=O)-[#16]-[#6]",
        ),  # N,N-ethyl thiocarbamate
        (range(7), "[#6]-[#16]-[#6](=O)-[#7](-[#6])-[#6]"),  # N,N-methyl thiocarbamate
    ]

    # Build list
    groups = []
    for add_if_match, template in group_templates:
        matches = mol.GetSubstructMatches(Chem.MolFromSmarts(template), useChirality=True)
        for match in matches:
            add_if = []
            for pattern_idx, atom_idx in enumerate(match):
                if pattern_idx in add_if_match:
                    add_if.append(atom_idx)
            groups.append((add_if, match))
    return groups


# Monkey patch RDChiral
rdchiral.template_extractor.get_special_groups = get_special_groups


def has_atom_mapping(smiles: str, is_smarts: bool = False, sanitize: bool = True) -> bool:
    """
    Returns True if a molecule has atom mapping, else False.

    :param smiles: the SMILES/SMARTS representing the molecule
    :param is_smarts: if True, will interpret the SMILES as a SMARTS
    :param sanitize: if True, will sanitize the molecule
    :return: True if the SMILES string has atom-mapping, else False
    """
    if not smiles:
        return False

    from_func = {
        False: functools.partial(Chem.MolFromSmiles, sanitize=sanitize),
        True: Chem.MolFromSmarts,
    }
    mol = from_func[is_smarts](smiles)
    if not mol:
        return False

    for atom in mol.GetAtoms():
        if atom.GetAtomMapNum() > 0:
            return True
    return False


def canonicalize_tautomer(smiles: str) -> str:
    """
    Returns the canonical tautomeric form of the input SMILES.
    """
    mol = Chem.MolFromSmiles(smiles)
    if not mol:
        return smiles
    tautomer_enumerator = rdMolStandardize.TautomerEnumerator()
    return Chem.MolToSmiles(tautomer_enumerator.Canonicalize(mol))


def enumerate_tautomers(smiles: str) -> List[str]:
    """
    Returns (sorted) collection of tautomers for the input SMILES.
    """
    mol = Chem.MolFromSmiles(smiles)
    tautomer_enumerator = rdMolStandardize.TautomerEnumerator()
    canonical_tautomer = tautomer_enumerator.Canonicalize(mol)

    canonical_tautomer_smiles = Chem.MolToSmiles(canonical_tautomer)
    tautomers = [canonical_tautomer_smiles]

    mols = tautomer_enumerator.Enumerate(mol)
    enumerated_smiles = [Chem.MolToSmiles(mol) for mol in mols if mol]

    enumerated_smiles = sorted(smiles for smiles in enumerated_smiles if smiles != canonical_tautomer_smiles)

    tautomers.extend(enumerated_smiles)
    return tautomers


def is_valid_mol(smiles: str) -> bool:
    """Check if the molecule structure is valid.

    :param smiles: Molecule in SMILES.
    :return: Return True if molecule structure is valid, return False otherwise.
    """
    mol = Chem.MolFromSmiles(smiles)
    if not mol:
        return False

    return True


def remove_stereochemistry(smiles: str) -> str:
    """Removing stereo-chemistry information from a SMILES."""
    mol = Chem.MolFromSmiles(smiles)
    if mol:
        return Chem.MolToSmiles(mol, isomericSmiles=False)
    return smiles


def remove_atom_mapping(smiles: str, is_smarts: bool = False, sanitize: bool = True, canonical: bool = True) -> str:
    """
    Returns a molecule without atom mapping

    :param smiles: the SMILES/SMARTS representing the molecule
    :param is_smarts: if True, will interpret the SMILES as a SMARTS
    :param sanitize: if True, will sanitize the molecule
    :param canonical: if False, will not canonicalize (applies to SMILES)
    :return: the molecule without atom-mapping
    """
    if not smiles:
        return ""

    # is_smarts=True: Chem.MolFromSmarts
    # is_smarts=False: Chem.MolFromSmiles
    from_func = {
        False: functools.partial(Chem.MolFromSmiles, sanitize=sanitize),
        True: Chem.MolFromSmarts,
    }
    # is_smarts=True: Chem.MolToSmarts
    # is_smarts=False: Chem.MolToSmiles
    to_func = {
        False: functools.partial(Chem.MolToSmiles, canonical=canonical),
        True: Chem.MolToSmarts,
    }
    mol = from_func[is_smarts](smiles)

    if not mol:
        return smiles

    for atom in mol.GetAtoms():
        atom.SetAtomMapNum(0)

    return to_func[is_smarts](mol)


def remove_atom_mapping_template(template_smarts: str) -> str:
    """Remove atom mapping from a template SMARTS string"""
    if not template_smarts:
        return ""

    rxn = AllChem.ReactionFromSmarts(template_smarts)
    rxn.Initialize()
    AllChem.RemoveMappingNumbersFromReactions(rxn)
    return AllChem.ReactionToSmarts(rxn)


def neutralize_molecules(smiles_list: List[str]) -> List[str]:
    """
    Neutralize a set of molecules using RDKit routines

    :param smiles_list: the molecules as SMILES
    :return: the neutralized molecules
    """
    # Neutralize/Uncharge Molecules
    uncharger = rdMolStandardize.Uncharger()
    # Create RDKit MolObjs
    mols = [Chem.MolFromSmiles(smi) for smi in smiles_list]
    # Neutralize Molecules
    neutral_mols = [uncharger.uncharge(mol) for mol in mols if mol]
    if len(mols) > len(neutral_mols):
        logging.warning(
            f"No. of Neutralised molecules ({len(neutral_mols)}) are less than No. of input molecules ({len(mols)})."
        )

    return [Chem.MolToSmiles(mol) for mol in neutral_mols]


def desalt_molecules(smiles_list: List[str], keep_something: bool = False) -> List[str]:
    """
    Remove salts from a set of molecules using RDKit routines

    :param smiles_list: the molecules as SMILES
    :param keep_something: if True will keep at least one salt
    :return: the desalted molecules
    """
    remover = SaltRemover()  # use default saltremover
    mols = [Chem.MolFromSmiles(smi) for smi in smiles_list]
    desalted_mols = [remover.StripMol(mol, dontRemoveEverything=keep_something) for mol in mols if mol]
    if len(mols) > len(desalted_mols):
        logging.warning(
            f"No. of desalted molecules ({len(desalted_mols)}) are less than No. of input molecules ({len(mols)})."
        )

    smiles_list_new = [Chem.MolToSmiles(mol) for mol in desalted_mols]
    smiles_list_new = [smi for smi in smiles_list_new if smi]
    if len(smiles_list) > len(smiles_list_new):
        logging.warning(
            f"No. of desalted SMILES ({len(smiles_list_new)}) are less than No. of input SMILES ({len(smiles_list)})."
        )
    return smiles_list_new


def same_molecule(mol1, mol2) -> bool:
    """Test if two molecules are the same.
    First number of atoms and bonds are compared to guard the potentially more expensive
    substructure match. If mol1 is a substructure of mol2 and vice versa, the molecules
    are considered to be the same.

    :param mol1: First molecule
    :param mol2: Second molecule for comparison
    :return: if the molecules match
    """
    return (
        mol1.GetNumHeavyAtoms() == mol2.GetNumHeavyAtoms()
        and (mol1.GetNumBonds() == mol2.GetNumBonds())
        and ((mol1.HasSubstructMatch(mol2)) and (mol2.HasSubstructMatch(mol1)))
    )


def atom_mapping_numbers(smiles: str) -> List[int]:
    """
    Return the numbers in the atom mapping

    :param smiles: the molecule as SMILES
    :return: the atom mapping numbers
    """
    mol = Chem.MolFromSmiles(smiles)
    if not mol:
        return []
    return [atom.GetAtomMapNum() for atom in mol.GetAtoms() if atom.GetAtomMapNum() > 0]


def reassign_rsmi_atom_mapping(rsmi: str, as_smiles: bool = False) -> str:
    """Reassign reaction's atom mapping.
    Remove atom maps for atoms in reactants and reactents not found in product's atoms.

    :param rsmi: Reaction SMILES
    :param as_smiles: Return reaction SMILES or SMARTS, defaults to False
    :return: Reaction SMILES or SMARTS
    """
    rxn = AllChem.ReactionFromSmarts(rsmi)
    logging.debug(f"Original rsmi: {rsmi}")
    logging.debug(f"Reactants: {rxn.GetNumReactantTemplates()}")
    logging.debug(f"Reagents: {rxn.GetNumAgentTemplates()}")
    logging.debug(f"Products: {rxn.GetNumProductTemplates()}")

    # Get Product(s) Atom Maps
    products_atommaps = {
        atom.GetAtomMapNum() for product in rxn.GetProducts() for atom in product.GetAtoms() if atom.GetAtomMapNum()
    }
    logging.debug(f"Product atom maps:{products_atommaps}")

    # Reassign Atom Maps for Rectants
    for reactant in rxn.GetReactants():
        reactant_atommaps = {
            atom.GetIdx(): atom.GetAtomMapNum() for atom in reactant.GetAtoms() if atom.GetAtomMapNum()
        }
        logging.debug(f"Reactant atom maps:{reactant_atommaps}")
        for atom_idx, atom_map in reactant_atommaps.items():
            # If atom map exists then continue
            if atom_map in products_atommaps:
                continue
            logging.debug(f"Atom {atom_idx} map num ({atom_map}) not found in product!!!")
            atom = reactant.GetAtomWithIdx(atom_idx)
            atom.SetAtomMapNum(0)
        logging.debug(f"Updated reactant: {Chem.MolToSmiles(reactant)}")
    # Reassign Atom Maps for Reagents
    for reagent in rxn.GetAgents():
        reagent_atommaps = {atom.GetIdx(): atom.GetAtomMapNum() for atom in reagent.GetAtoms() if atom.GetAtomMapNum()}
        logging.debug(f"Reagent atom maps: {reagent_atommaps}")
        for atom_idx, atom_map in reagent_atommaps.items():
            # If atom map exists then continue
            if atom_map in products_atommaps:
                continue
            logging.debug(f"Atom {atom_idx} map num ({atom_map}) not found in product!!!")
            atom = reagent.GetAtomWithIdx(atom_idx)
            atom.SetAtomMapNum(0)
        logging.debug(f"Updated reagent: {Chem.MolToSmiles(reagent)}")
    # Get updated rsmi
    updated_rsmi = AllChem.ReactionToSmiles(rxn) if as_smiles else AllChem.ReactionToSmarts(rxn)
    logging.debug(f"Updated rsmi: {updated_rsmi}")

    return updated_rsmi


def split_rsmi(rsmi: str) -> Tuple[str, str, str]:
    """
    Split a reaction SMILES into components SMILES

    :param rsmi: the reaction SMILES
    :return: the SMILES of the components
    """
    reaction_components = rsmi_split_pattern.split(rsmi)
    num_reaction_components = len(reaction_components)
    if num_reaction_components != 3:
        raise ValueError(f"Expected 3 reaction components but got {num_reaction_components} for '{rsmi}'")

    return tuple(reaction_components)


def join_smiles_from_reaction(smiles_list: List[str]) -> str:
    """
    Join a part of reaction SMILES, e.g. reactants and products into components.
    Intra-molecular complexes are bracketed with parenthesis

    :param smiles_list: the SMILES components
    :return: the joined list
    """
    return ".".join([f"({item})" if "." in item else item for item in smiles_list])


def split_smiles_from_reaction(smiles: str) -> List[str]:
    """
    Split a part of reaction SMILES, e.g. reactants or products
    into components. Taking care of intra-molecular complexes

    Taken from RDKit:
    https://github.com/rdkit/rdkit/blob/master/Code/GraphMol/ChemReactions/DaylightParser.cpp

    :param smiles: the SMILES/SMARTS
    :return: the individual components.
    """
    pos = 0
    block_start = 0
    level = 0
    in_block = 0
    components = []
    while pos < len(smiles):
        if smiles[pos] == "(":
            if pos == block_start:
                in_block = 1
            level += 1
        elif smiles[pos] == ")":
            if level == 1 and in_block:
                in_block = 2
            level -= 1
        elif level == 0 and smiles[pos] == ".":
            if in_block == 2:
                components.append(smiles[block_start + 1 : pos - 1])
            else:
                components.append(smiles[block_start:pos])
            block_start = pos + 1
            in_block = 0
        pos += 1
    if block_start < pos:
        if in_block == 2:
            components.append(smiles[block_start + 1 : pos - 1])
        else:
            components.append(smiles[block_start:pos])
    return components


def recreate_rsmi(rsmi: str) -> str:
    """
    Recreate Reactions SMILES by removing intra-molecular complexes.

    :param rsmi: the original reaction smiles
    :return: the updated reaction smiles without intra-molecular complexes
    """
    reactants, reagents, products = split_rsmi(rsmi)
    # Split and rejoin the components
    reactants = ".".join(split_smiles_from_reaction(reactants))
    reagents = ".".join(split_smiles_from_reaction(reagents))
    products = ".".join(split_smiles_from_reaction(products))

    new_rsmi = f"{reactants}>{reagents}>{products}"

    return new_rsmi


def reaction_centres(rxn: AllChem.ChemicalReaction) -> Tuple[List[int], ...]:
    """
    Return reaction centre atoms, provided that the bonding partners
    actually change when comparing the environment in the reactant and the product

    inspired by code from Greg Landrum's tutorial
    set up array to remove atoms from the reaction centers
    by comparing the atom mapping in the reactant vs the products

    Original implementation by Christoph Bauer

    :param rxn: the initialized RDKit reaction
    :return: tuple of reaction centre atoms, filtered by connectivity criterion
    """
    rxncenters = rxn.GetReactingAtoms(mappedAtomsOnly=True)

    remove_array = []
    nreactants = len(rxncenters)
    for ridx, reacting in enumerate(rxncenters):
        reactant = rxn.GetReactantTemplate(ridx)
        for raidx in reacting:
            ratm = reactant.GetAtomWithIdx(raidx)
            mapnum = ratm.GetAtomMapNum()
            numexplicitH_reactant = ratm.GetNumExplicitHs()
            neighbours = ratm.GetNeighbors()
            map_neighbours = sorted([neighbour_atom.GetAtomMapNum() for neighbour_atom in neighbours])
            foundit = False
            for product in rxn.GetProducts():
                for patom in product.GetAtoms():
                    if patom.GetAtomMapNum() == mapnum:
                        neighbours_product = patom.GetNeighbors()
                        numexplicitH_product = patom.GetNumExplicitHs()
                        map_neighbours_product = sorted(
                            [neighbour_atom.GetAtomMapNum() for neighbour_atom in neighbours_product]
                        )
                        # criterion: check if the map numbers of neighbours are the same in the reactant and the product
                        if map_neighbours == map_neighbours_product:
                            if numexplicitH_reactant == numexplicitH_product:
                                # if yes --> then the environment doesn't change, so set remove to True
                                remove_array.append(True)
                            else:
                                remove_array.append(False)
                        else:
                            # False means that the environment does change, so set remove to False
                            remove_array.append(False)
                        foundit = True
                        break
                    if foundit:
                        break

    # actual removal of reaction centres by generating new tuple
    rxncenters_filtered = []
    counter = 0
    for reactant_idx in range(nreactants):
        reactantcenters = []
        for index in rxncenters[reactant_idx]:
            if not remove_array[counter]:
                reactantcenters.append(index)
            counter += 1
        reactantcenters = tuple(reactantcenters)
        rxncenters_filtered.append(reactantcenters)
    rxncenters_filtered = tuple(rxncenters_filtered)
    return rxncenters_filtered



""" Contains routines for computing route similarities
"""

import functools
from typing import Any, Callable, Dict, List, Sequence, Set, Tuple

import numpy as np

from rxnutils.chem.reaction import ChemicalReaction
from rxnutils.chem.utils import atom_mapping_numbers, split_rsmi
from rxnutils.routes.base import SynthesisRoute
from rxnutils.routes.ted.distances_calculator import ted_distances_calculator

RouteDistancesCalculator = Callable[[Sequence[SynthesisRoute]], np.ndarray]


def simple_route_similarity(routes: Sequence[SynthesisRoute]) -> np.ndarray:
    """
    Returns the geometric mean of the simple bond forming similarity, and
    the atom matching bonanza similarity

    :param routes: the sequence of routes to compare
    :return: the pairwise similarity
    """
    bond_score = simple_bond_forming_similarity(routes)
    atom_score = atom_matching_bonanza_similarity(routes)
    return np.sqrt(bond_score * atom_score)


def atom_matching_bonanza_similarity(routes: Sequence[SynthesisRoute]) -> np.ndarray:
    """
    Calculates the pairwise similarity of a sequence of routes
    based on the overlap of the atom-mapping numbers of the compounds
    in the routes.

    :param routes: the sequence of routes to compare
    :return: the pairwise similarity
    """
    sims = np.zeros((len(routes), len(routes)))
    for idx1, route1 in enumerate(routes):
        mols1 = _extract_atom_mapping_numbers(route1)
        n1 = len(mols1) - 1
        mols1 = [mol for mol in mols1 if mol]
        for idx2, route2 in enumerate(routes):
            if idx1 == idx2:
                sims[idx1, idx2] = 1.0
                continue
            # If both of the routes have not reactions, we assume
            # they are identical and it is assignes a score of 1.0
            if route1.max_depth == 0 and route2.max_depth == 0:
                sims[idx1, idx2] = 1.0
                continue
            # If just one the route has no reactions, we assume
            # they are maximal dissimilar and assigns a score 0.0 (see how `sims` is initialized)
            if route1.max_depth == 0 or route2.max_depth == 0:
                continue

            mols2 = _extract_atom_mapping_numbers(route2)
            n2 = len(mols2) - 1
            mols2 = [mol for mol in mols2 if mol]

            o = _calc_overlap_matrix(mols1[1:], mols2[1:])
            sims[idx1, idx2] = (o.max(axis=1).sum() + o.max(axis=0).sum()) / (n1 + n2)
    return sims


def simple_bond_forming_similarity(routes: Sequence[SynthesisRoute]) -> np.ndarray:
    """
    Calculates the pairwise similarity of a sequence of routes
    based on the overlap of formed bonds in the reactions.

    :param routes: the sequence of routes to compare
    :return: the pairwise similarity
    """
    sims = np.ones((len(routes), len(routes)))
    for idx1, route1 in enumerate(routes):
        bonds1 = _extract_formed_bonds(route1)
        for idx2, route2 in enumerate(routes):
            if idx2 <= idx1:
                continue
            # If both of the routes have not reactions, we assume
            # they are identical and it is assignes a score of 1.0 (see how `sims` is initialized)
            if route1.max_depth == 0 and route2.max_depth == 0:
                continue
            # If just one the route has no reactions, we assume
            # they are maximal dissimilar and assigns a score 0.0
            if route1.max_depth == 0 or route2.max_depth == 0:
                sims[idx1, idx2] = sims[idx2, idx1] = 0.0
                continue

            bonds2 = _extract_formed_bonds(route2)
            sims[idx1, idx2] = _bond_formed_overlap_score(bonds1, bonds2)
            sims[idx2, idx1] = sims[idx1, idx2]  # Score is symmetric
    return sims


def route_distances_calculator(model: str, **kwargs: Any) -> RouteDistancesCalculator:
    """
    Return a callable that given a list routes as dictionaries
    calculate the squared distance matrix

    :param model: the route distance model name
    :param kwargs: additional keyword arguments for the model
    :return: the appropriate route distances calculator
    """
    if model not in ["ted", "lstm"]:
        raise ValueError("Model must be either 'ted' or 'lstm'")

    if model == "ted":
        model_kwargs = _copy_kwargs(["content", "timeout"], **kwargs)
        return functools.partial(ted_distances_calculator, **model_kwargs)

    # Placeholder for LSTM distances calculation
    # model_kwargs = _copy_kwargs(["model_path"], **kwargs)
    # return lstm_distances_calculator(**model_kwargs)
    raise NotImplementedError("LSTM route distances calculator not implemented yet.")


def _copy_kwargs(keys_to_copy: List[str], **kwargs: Any) -> Dict[str, Any]:
    """Copy selected keyword arguments."""
    new_kwargs = {}
    for key in keys_to_copy:
        if key in kwargs:
            new_kwargs[key] = kwargs[key]
    return new_kwargs


def _calc_overlap_matrix(mols1: List[List[int]], mols2: List[List[int]]) -> np.ndarray:
    """
    Calculate the pairwise overlap matrix between
    the molecules in the two input vectors

    :param mols1: the atom-mapping numbers of the first molecule
    :param mols2: the atom-mapping numbers of the second molecule
    :return: the computed matrix
    """

    def mol_overlap(mol1, mol2):
        if not mol1 and not mol2:
            return 0
        return len(set(mol1).intersection(mol2)) / max(len(mol1), len(mol2))

    overlaps = np.zeros((len(mols1), len(mols2)))
    for idx1, mol1 in enumerate(mols1):
        for idx2, mol2 in enumerate(mols2):
            overlaps[idx1, idx2] = mol_overlap(mol1, mol2)
    return overlaps


def _extract_atom_mapping_numbers(route: SynthesisRoute) -> List[List[int]]:
    """
    Extract all the compounds in a synthesis routes as a set of
    atom-mapping numbers of the atoms in the compounds.

    Only account for atom-mapping numbers that exists in the root compound
    """
    if route.max_depth == 0:
        return []

    root_atom_numbers = sorted(atom_mapping_numbers(route.mapped_root_smiles))
    mapping_list = [root_atom_numbers]
    for reaction_smiles in route.atom_mapped_reaction_smiles():
        reactants_smiles = split_rsmi(reaction_smiles)[0]
        for smi in reactants_smiles.split("."):
            atom_numbers = sorted([num for num in atom_mapping_numbers(smi) if num in root_atom_numbers])
            mapping_list.append(atom_numbers)
    return mapping_list


def _extract_formed_bonds(route: SynthesisRoute) -> List[Tuple[int, int]]:
    """
    Extract a set of bonds formed in the synthesis routes.
    Only bonds contributing to the root compound is accounted for.
    """
    if route.max_depth == 0:
        return []

    formed_bonds = []
    root_atom_numbers = set(atom_mapping_numbers(route.mapped_root_smiles))
    for reaction_smiles in route.atom_mapped_reaction_smiles():
        rxn_obj = ChemicalReaction(reaction_smiles, clean_smiles=False)
        product_bonds = set()
        product_bonds = _extract_molecule_bond_tuple(rxn_obj.products)
        reactants_bonds = _extract_molecule_bond_tuple(rxn_obj.reactants)
        for idx1, idx2 in product_bonds - reactants_bonds:
            if idx1 in root_atom_numbers and idx2 in root_atom_numbers:
                formed_bonds.append(tuple(sorted([idx1, idx2])))
    return formed_bonds


def _extract_molecule_bond_tuple(molecules: List) -> Set[int]:
    """
    Returns the sorted set of atom-mapping number pairs for each
    bonds in a list of rdkit molecules
    """
    bonds = set()
    for bond in [bond for mol in molecules for bond in mol.GetBonds()]:
        bonds.add(
            tuple(
                sorted(
                    [
                        bond.GetBeginAtom().GetAtomMapNum(),
                        bond.GetEndAtom().GetAtomMapNum(),
                    ]
                )
            )
        )
    return bonds


def _bond_formed_overlap_score(bonds1: List[Tuple[int]], bonds2: List[Tuple[int]]) -> float:
    """
    Computes a similarity score of two routes by comparing the overlap
    of bonds formed in the synthesis routes.
    """
    overlap = len(set(bonds1).intersection(bonds2))
    if not bonds1 and not bonds2:
        return 0
    return overlap / max(len(bonds1), len(bonds2))



""" Module with featurizers for molecules and reactions
"""

from operator import itemgetter
from typing import Any, Callable, Dict, List, Tuple

import numpy as np
from rdkit.Chem import AllChem, DataStructs

from rxnutils.chem.reaction import ChemicalReaction


def collect_reaction_features(
    reaction_data: List[Dict[str, Any]],
    target_fingerprint: np.ndarray,
    reaction_class_ranks: Dict[str, int],
    featurizer: Callable[[Dict[str, Any]], np.ndarray],
    prefactor: float = 1.1,
    default_rank: int = 5,
    default_cls: str = "0.0",
) -> Tuple[float, np.ndarray]:
    """
    Collect all reaction features and compute the prior reaction feasibility score

    :params reaction_data: the reactions of the route
    :params target_fingerprint: the precomputed fingerprint of the target
    :params reaction_class_ranks: the ranks of the reaction classes
    :params featurizer: the featurizer to use for reactions
    :params prefactor: the factor for the feasibility score
    :params default_rank: the rank of reactions not parameterized
    :params default_cls: the class of reactions not parameterized
    :returns: the prior feasibility score
    :returns: the reaction features
    """
    weights = [prefactor**idx for idx in range(1, len(reaction_data) + 1)]
    weights_sum = sum(weights)

    classes = []
    ranks = []
    fingerprints = []
    rank_score = 0.0
    # This result in reactions being traversed breadth-first instead of depth-first
    reaction_data_sorted = sorted(reaction_data, key=itemgetter("tree_depth"))
    for weight, reaction_data in zip(weights, reaction_data_sorted):
        norm_weight = weight / weights_sum
        cls = reaction_data.get("classification", default_cls)
        if " " in cls:
            cls = cls.split(" ")[0]
        if cls in reaction_class_ranks:
            rank = reaction_class_ranks[cls]
        else:
            rank = default_rank
            cls = default_cls
        rank_score += norm_weight * rank

        class_vector = [int(item) for item in cls.split(".")]
        class_vector += [0] * (3 - len(class_vector))
        classes.append(class_vector)
        ranks.append(rank)
        fingerprints.append(featurizer(reaction_data))

    features = []
    for class_, rank, fingerprint in zip(classes, ranks, fingerprints):
        vector = np.asarray(class_ + [rank])
        vector = np.concatenate([vector, target_fingerprint, fingerprint])
        features.append(vector)
    return rank_score, np.asarray(features)


def default_reaction_featurizer(reaction_data: Dict[str, Any], radius: int = 2, numbits: int = 64) -> np.ndarray:
    """
    Given an item of reaction data, returned by the `reaction_data` method
    of `SynthesisRoute` this features the reaction using a difference ECFP

    Currently it featurizes the reaction in the retro-sense

    :params reaction_data: the reaction as a dictionary
    :params radius: the radius of the fingerprint
    :params numbits: the number of bits
    :returns: the fingerprint
    """
    rsmi = ">>".join(reaction_data["reaction_smiles"].split(">>")[::-1])
    rxn = ChemicalReaction(rsmi, clean_smiles=False)
    return reaction_difference_fingerprint(rxn, radius, numbits)


def ecfp_fingerprint(mol: AllChem.rdchem.Mol, radius: int = 2, numbits: int = 64) -> np.ndarray:
    """
    Computes a ECFP with a given radius and length

    :params mol: the RDKit molecule
    :params radius: the radius of the fingerprint
    :params numbits: the number of bits
    :returns: the fingerprint
    """
    array = np.zeros((0,), dtype=np.int8)
    fingerprint = AllChem.GetMorganFingerprintAsBitVect(mol, radius, nBits=numbits)
    DataStructs.ConvertToNumpyArray(fingerprint, array)
    return array


def reaction_difference_fingerprint(reaction: ChemicalReaction, radius: int = 2, numbits: int = 64) -> np.ndarray:
    """
    Computes the difference fingerprint of given reaction

    :params reaction: the reaction to featurize
    :params radius: the radius of the fingerprint
    :params numbits: the number of bits
    :returns: the fingerprint
    """
    fp = np.zeros(numbits)
    for product in reaction.products:
        if product is not None:
            fp += ecfp_fingerprint(product, radius, numbits)
    for reactant in reaction.reactants:
        if reactant is not None:
            fp -= ecfp_fingerprint(reactant, radius, numbits)
    return fp



"""Module containing base class for data pipelines
"""

from pathlib import Path
from typing import List, Tuple

from metaflow import FlowSpec, Parameter

from rxnutils.data.batch_utils import combine_csv_batches, create_csv_batches

# This is hack to only import the validation_runner if rxnmapper is not installed
try:
    import rxnmapper  # noqa
except ImportError:
    from rxnutils.pipeline.runner import main as validation_runner


class DataBaseFlow(FlowSpec):
    """Base-class for pipelines for processing data"""

    nbatches = Parameter("nbatches", type=int, required=True)
    folder = Parameter("folder", default=".")

    def _combine_batches(self, filename: str) -> None:
        combine_csv_batches(filename, self.nbatches)

    def _create_batches(self, input_filename: str, output_filename: str) -> List[Tuple[int, int, int]]:
        return create_csv_batches(
            input_filename,
            self.nbatches,
            output_filename,
        )


class DataPreparationBaseFlow(DataBaseFlow):
    """Base pipeline for preparing datasets and doing clean-up"""

    data_prefix = ""

    def _setup_cleaning(self):
        """Setup cleaning"""
        # pylint: disable=attribute-defined-outside-init
        self.partitions = self._create_batches(
            Path(self.folder) / f"{self.data_prefix}_data.csv",
            Path(self.folder) / f"{self.data_prefix}_data_cleaned.csv",
        )

    def _do_cleaning(self):
        """Perform cleaning of data"""
        idx, start, end = self.input
        pipeline_path = str(Path(__file__).parent / "clean_pipeline.yml")
        if idx > -1:
            validation_runner(
                [
                    "--pipeline",
                    pipeline_path,
                    "--data",
                    str(Path(self.folder) / f"{self.data_prefix}_data.csv"),
                    "--output",
                    str(Path(self.folder) / f"{self.data_prefix}_data_cleaned.csv.{idx}"),
                    "--batch",
                    str(start),
                    str(end),
                    "--max-workers",
                    "1",
                    "--no-intermediates",
                ]
            )

    def _join_cleaning(self):
        """Combined cleaned batches of data"""
        self._combine_batches(Path(self.folder) / f"{self.data_prefix}_data_cleaned.csv")



"""
Contains a class encapsulating a synthesis route,
as well as routines for assigning proper atom-mapping
and drawing the route
"""

import warnings
from collections import defaultdict
from copy import deepcopy
from operator import itemgetter
from typing import Any, Callable, Dict, List, Set, Tuple, Union

import pandas as pd
from PIL.Image import Image as PilImage
from rdkit import Chem

from rxnutils.chem.augmentation import single_reactant_augmentation
from rxnutils.chem.utils import atom_mapping_numbers, join_smiles_from_reaction, split_rsmi, split_smiles_from_reaction
from rxnutils.pipeline.actions.reaction_mod import NameRxn, RxnMapper
from rxnutils.routes.image import RouteImageFactory
from rxnutils.routes.utils.validation import validate_dict


class SynthesisRoute:
    """
    This encapsulates a synthesis route or a reaction tree.
    It provide convinient methods for assigning atom-mapping
    to the reactions, and for providing reaction-level data
    of the route

    It is typically initiallized by one of the readers in the
    `rxnutils.routes.readers` module.

    The tree depth and the forward step is automatically assigned
    to each reaction node.

    The `max_depth` attribute holds the longest-linear-sequence (LLS)

    :param reaction_tree: the tree structure representing the route
    """

    def __init__(self, reaction_tree: Dict[str, Any]) -> None:
        validate_dict(reaction_tree)
        self.reaction_tree = reaction_tree
        self.max_depth = _assign_tree_depth(reaction_tree)
        _assign_forward_step(reaction_tree, self.max_depth)
        self._nsteps = -1

    @property
    def mapped_root_smiles(self) -> str:
        """
        Return the atom-mapped SMILES of the root compound

        Will raise an exception if the route is a just a single
        compound, or if the route has not been assigned atom-mapping.
        """
        if len(self.reaction_tree.get("children", [])) == 0:
            raise ValueError("Single-compound root does not have this property")

        first_reaction = self.reaction_tree["children"][0]
        if "mapped_reaction_smiles" not in first_reaction.get("metadata", {}):
            raise ValueError("It appears that the route has no atom-mapping information")

        return split_rsmi(first_reaction["metadata"]["mapped_reaction_smiles"])[-1]

    @property
    def nsteps(self) -> int:
        """Return the number of reactions in the route"""
        if self._nsteps == -1:
            self._nsteps = len(self.reaction_smiles())
        return self._nsteps

    def atom_mapped_reaction_smiles(self) -> List[str]:
        """Returns a list of the atom-mapped reaction SMILES in the route"""
        smiles = []
        _collect_atom_mapped_smiles(self.reaction_tree, smiles)
        return smiles

    def assign_atom_mapping(self, overwrite: bool = False, only_rxnmapper: bool = False) -> None:
        """
        Assign atom-mapping to each reaction in the route and
        ensure that is is consistent from root compound and throughout
        the route.

        It will use NameRxn to assign classification and possiblty atom-mapping,
        as well as rxnmapper to assign atom-mapping in case NameRxn cannot classify
        a reaction.

        :param overwrite: if True will overwrite existing mapping
        :param only_rxnmapper: if True will disregard NameRxn mapping and use only rxnmapper
        """
        self._assign_mapping(overwrite, only_rxnmapper)
        max_atomnum = max(atom_mapping_numbers(self.mapped_root_smiles))
        _inherit_atom_mapping(self.reaction_tree, max_atomnum + 100)

    def chains(self, complexity_func: Callable[[str], float]) -> List[List[Dict[str, Any]]]:
        """
        Returns linear sequences or chains extracted from the route.

        Each chain is a list of a dictionary representing the molecules, only the most
        complex molecule is kept for each reaction - making the chain a sequence of molecule
        to molecule transformation.

        The first chain will be the longest linear sequence (LLS), and the second chain
        will be longest branch if this is a convergent route. This branch will be processed
        further, but the other branches can probably be discarded as they have not been
        investigated thoroughly.

        :param complexity_func: a function that takes a SMILES and returns a
                            complexity metric of the molecule
        :returns: a list of chains where each chain is a list of molecules
        """
        mols = []
        _extract_chain_molecules(self.reaction_tree, mols)
        chains = _extract_chains(mols, complexity_func)

        # Process the longest branch for convergent routes
        # note that other chains will not be processed
        # and thus should probably be discarded
        # TODO: analyse other chains
        if len(chains) > 1:
            last_step = chains[1][-1]["step"]
            mol_copy = dict(chains[0][last_step + 1])
            mol_copy["serial"] += "x"
            chains[1].append(mol_copy)

        # Add name of chain and type of molecule to each molecule
        for idx, chain in enumerate(chains):
            chain_name = "lls" if idx == 0 else f"sub{idx}"
            for mol in chain:
                mol["chain"] = chain_name
                if mol["parent_hash"] is None:
                    mol["type"] = "target"
                elif "hash" not in mol:
                    mol["type"] = "sm"
                else:
                    mol["type"] = "inter"

                if mol["serial"].endswith("x"):
                    mol["type"] = "branch"
                    mol["chain"] = "sub1"
        return chains

    def image(self, show_atom_mapping: bool = False, factory_kwargs: Dict[str, Any] = None) -> PilImage:
        """
        Depict the route.

        :param show_atom_mapping: if True, will show the atom-mapping
        :param factory_kwargs: additional keyword arguments sent to the `RouteImageFactory`
        :returns: the image of the route
        """
        factory_kwargs = factory_kwargs or {}
        if show_atom_mapping:
            dict_ = self._create_atom_mapping_tree()
        else:
            dict_ = self.reaction_tree
        return RouteImageFactory(dict_, **factory_kwargs).image

    def intermediate_counts(self) -> Dict[str, int]:
        """
        Extract the counts of all intermediates

        return: the counts
        """
        intermediates = defaultdict(int)
        children = self.reaction_tree.get("children", [])
        if children:
            _collect_intermediates(children[0], intermediates)
        return intermediates

    def intermediates(
        self,
    ) -> Set[str]:
        """
        Extract a set with the SMILES of all the intermediates nodes

        :return: a set of SMILES strings
        """
        return set(self.intermediate_counts().keys())

    def is_solved(self) -> bool:
        """
        Find if this route is solved, i.e. if all starting material
        is in stock.

        To be accurate, each molecule node need to have an extra
        boolean property called `in_stock`.
        """
        try:
            _find_leaves_not_in_stock(self.reaction_tree)
        except ValueError:
            return False
        return True

    def leaf_counts(self) -> Dict[str, int]:
        """
        Extract the counts of all leaf nodes, i.e. starting material

        return: the counts
        """
        leaves = defaultdict(int)
        _collect_leaves(self.reaction_tree, leaves)
        return leaves

    def leaves(
        self,
    ) -> Set[str]:
        """
        Extract a set with the SMILES of all the leaf nodes, i.e.
        starting material

        :return: a set of SMILES strings
        """
        return set(self.leaf_counts().keys())

    def reaction_data(self) -> List[Dict[str, Any]]:
        """
        Returns a list of dictionaries for each reaction
        in the route. This is metadata of the reactions
        augmented with reaction SMILES and depth of the reaction
        """
        data = []
        _collect_reaction_data(self.reaction_tree, data)
        return data

    def reaction_ngrams(self, nitems: int, metadata_key: str) -> List[Tuple[Any, ...]]:
        """
        Extract an n-gram representation of the route by building up n-grams
        of the reaction metadata.

        :param nitems: the length of the gram
        :param metadata_key: the metadata to extract
        :return: the collected n-grams
        """
        if self.max_depth < nitems:
            return []
        first_reaction = self.reaction_tree["children"][0]
        ngrams = []
        _collect_ngrams(first_reaction, nitems, metadata_key, ngrams)
        return ngrams

    def reaction_smiles(self, augment: bool = False) -> List[str]:
        """
        Returns a list of the un-mapped reaction SMILES
        :param augment: if True will add reagents to single-reactant
                          reagents whenever possible
        """
        if not augment:
            smiles_list = []
            _collect_reaction_smiles(self.reaction_tree, smiles_list)
            return smiles_list

        smiles_list = []
        for data in self.reaction_data():
            smiles = data["reaction_smiles"]
            classification = data.get("classification", "")
            smiles_list.append(single_reactant_augmentation(smiles, classification))
        return smiles_list

    def remap(self, other: Union["SynthesisRoute", str, Dict[int, int]]) -> None:
        """
        Remap the reaction so that it follows the mapping of a
        1) root compound in a reference route, 2) a ref compound given
        as a SMILES, or 3) using a raw mapping

        :param other: the reference for re-mapping
        """
        if isinstance(other, SynthesisRoute):
            if len(self.reaction_smiles()) == 0 or len(other.reaction_smiles()) == 0:
                return
            mapping_dict = _find_remapping(other.mapped_root_smiles, self.mapped_root_smiles)
        elif isinstance(other, str):
            if len(self.reaction_smiles()) == 0:
                return
            mapping_dict = _find_remapping(other, self.mapped_root_smiles)
        elif isinstance(other, dict):
            mapping_dict = other
        else:
            raise ValueError(f"Cannot perform re-mapping using a {type(other)}")
        _remap_reactions(self.reaction_tree, mapping_dict)

    def _assign_mapping(self, overwrite: bool = False, only_rxnmapper: bool = False) -> None:
        if not overwrite:
            try:
                self.atom_mapped_reaction_smiles()
            except KeyError:
                # We will just create the mapping
                pass
            else:
                return

        df = pd.DataFrame({"smiles": list(set(self.reaction_smiles()))})
        nextmove_action = NameRxn(in_column="smiles", nm_rxn_column="mapped_smiles")
        rxnmapper_action = RxnMapper(in_column="smiles")
        if not only_rxnmapper:
            try:
                df = nextmove_action(df)
            # Raised by nextmove_action if namerxn not in path
            except FileNotFoundError:
                df = df.assign(NMC=["0.0"] * len(df), mapped_smiles=[""] * len(df))
                warnings.warn("namerxn does not appear to be in $PATH. Run failed and proceeding with rxnmapper only")
        else:
            df = df.assign(NMC=["0.0"] * len(df), mapped_smiles=[""] * len(df))
        df = rxnmapper_action(df)
        if only_rxnmapper:
            df["mapped_smiles"] = df["RxnmapperRxnSmiles"]
        else:
            sel = df["NMC"] == "0.0"
            df.loc[sel, "mapped_smiles"] = df.loc[sel, "RxnmapperRxnSmiles"]
        datamap = df.set_index("smiles").to_dict("index")
        _copy_mapping_from_datamap(self.reaction_tree, datamap)

    def _create_atom_mapping_tree(self) -> Dict[str, Any]:
        dict_ = deepcopy(self.reaction_tree)
        _assign_atom_mapped_smiles(dict_)
        return dict_


def _apply_remapping(
    reaction_smiles: str,
    remapping: Dict[int, int],
    keep_original: bool,
) -> str:
    """
    Apply a specific re-mapping to a given reaction

    If an atom has an atom-mapping number not given
    by the re-mapping dictionary, the `keep_original` argument
    determines the action. If `keep_original` is True, the atom
    keeps the atom-mapping number, else it is set to 0 and
    becomes and un-mapped atom.

    :param reaction_smiles: the reaction to remap
    :param remapping: the mapping from old to new atom-map number
    :param keep_original: determines the behavior of atoms not in `remapping`
    :return: the remapped reaction SMILES
    """
    reactants_smiles, reagent_smiles, product_smiles = split_rsmi(reaction_smiles)
    product_mol = Chem.MolFromSmiles(product_smiles)
    reactant_mols = [Chem.MolFromSmiles(smiles) for smiles in split_smiles_from_reaction(reactants_smiles)]
    atoms_to_renumber = list(product_mol.GetAtoms())
    for mol in reactant_mols:
        atoms_to_renumber.extend(mol.GetAtoms())
    for atom in atoms_to_renumber:
        if atom.GetAtomMapNum() and atom.GetAtomMapNum() in remapping:
            atom.SetAtomMapNum(remapping[atom.GetAtomMapNum()])
        elif not keep_original:
            atom.SetAtomMapNum(0)
    return ">".join(
        [
            join_smiles_from_reaction([Chem.MolToSmiles(mol) for mol in reactant_mols]),
            reagent_smiles,
            Chem.MolToSmiles(product_mol),
        ]
    )


def _extract_chains(
    chain_molecules: List[Dict[str, Any]], complexity_func: Callable[[str], float]
) -> List[List[Dict[str, Any]]]:
    """
    Take extracted molecules from a route and compile them into chains or sequences.

    The molecules will first be sorted based on complexity and then by step.
    Therefore the chains will only contain the most complex reactant of a reaction,
    and the first chain will be the longest linear sequence (LLS)

    Single-molecule chains will be discarded.

    :param chain_molecules: the list of molecules
    :param complexity_func: a function that takes a SMILES and returns a
                            complexity metric of the molecule
    :returns: a list of chains where each chain is a list of molecules
    """
    # Add complexity and serial number to each molecule
    for idx, mol in enumerate(chain_molecules):
        mol["complexity"] = complexity_func(mol["smiles"])
        mol["serial"] = str(len(chain_molecules) - idx - 1)
    chain_molecules.sort(key=itemgetter("complexity"), reverse=True)
    chain_molecules.sort(key=itemgetter("step"))

    chains = []
    residuals = list(chain_molecules)
    while len(residuals) > 0:
        if len(residuals) == 1:
            chains.append([residuals[0]])
            break

        # Start a new chain with the first residual molecule
        # then find the linear sequence from that molecule
        new_chain = [residuals[0]]
        residuals = residuals[1:]
        # Loop until we don't find a parent molecule
        found = True
        while found:
            found = False
            # Look for a parent molecule in the residuals
            for mol in residuals:
                if "hash" in mol and mol["hash"] == new_chain[-1]["parent_hash"]:
                    new_chain.append(mol)
                    found = True
                    break
        residuals = [mol for mol in residuals if not mol in new_chain]
        chains.append(new_chain)

    chains.sort(key=len, reverse=True)
    chains = [chain for chain in chains if len(chain) > 1]
    return chains


def _find_remapping(parent_smiles: str, target_smiles: str) -> Dict[int, int]:
    """
    Find the mapping between two identical copies of a compound
    that have different atom-mapping.

    :param parent_smiles: the reference SMILES for which the mapping should be changed to
    :param target_smiles: the other SMILES that will be re-mapped
    :return: the translation of atom mapping from `product_smiles` to `parent_smiles`
    """
    remapping = {}
    pmol = Chem.MolFromSmiles(parent_smiles)
    tmol = Chem.MolFromSmiles(target_smiles)

    for atom_idx1, atom_idx2 in enumerate(pmol.GetSubstructMatch(tmol)):
        atom1 = tmol.GetAtomWithIdx(atom_idx1)
        atom2 = pmol.GetAtomWithIdx(atom_idx2)
        if atom1.GetAtomMapNum() > 0 and atom2.GetAtomMapNum() > 0:
            remapping[atom1.GetAtomMapNum()] = atom2.GetAtomMapNum()
    return remapping


def smiles2inchikey(smiles: str, ignore_stereo: bool = False) -> str:
    """Converts a SMILES to an InChI key"""
    inchi = Chem.MolToInchiKey(Chem.MolFromSmiles(smiles))
    if not ignore_stereo:
        return inchi
    return inchi.split("-")[0]


# Recursive functions acting on reaction trees


def _assign_atom_mapped_smiles(tree_dict: Dict[str, Any], reactants_smiles: List[str] = None) -> None:
    """
    Used to copy the atom-mapped SMILES from the reaction metadata
    to the SMILES property of each molecule node. This is used to
    display atom-mapping when depicting routes.

    The matching of atom-mapped SMILES and unmapped SMILES are
    based on partial InChI key without stereoinformation

    :param tree_dict: the current molecule node
    :param reactants_smiles: the list of reactants SMILES of the parent reaction
    """
    tree_dict["unmapped_smiles"] = tree_dict["smiles"]
    # For leaf nodes
    if not tree_dict.get("children", []):
        inchi = smiles2inchikey(tree_dict["smiles"], ignore_stereo=True)
        found = -1
        for idx, smi in enumerate(reactants_smiles):
            if inchi == smiles2inchikey(smi, ignore_stereo=True):
                found = idx
                break
        if found > -1:
            tree_dict["smiles"] = reactants_smiles.pop(found)
    else:
        reaction_dict = tree_dict["children"][0]
        reaction_smiles = reaction_dict["metadata"]["mapped_reaction_smiles"]
        reactants, _, tree_dict["smiles"] = split_rsmi(reaction_smiles)
        reactants_smiles = split_smiles_from_reaction(reactants)
        for grandchild in reaction_dict["children"]:
            _assign_atom_mapped_smiles(grandchild, reactants_smiles)


def _assign_forward_step(tree_dict: Dict[str, Any], max_depth: int, depth: int = 1) -> None:
    """
    Assign the forward_step property of each reaction node.
    The forward step is defined as "maximum depth - node depth + 1"

    :param tree_dict: the current molecule node
    :param max_depth: the maximum depth of the full tree
    :param depth: the depth of the current molecule node + 1
    """
    children = tree_dict.get("children")
    if children is None:
        return
    grandchildren = children[0]["children"]
    children[0]["metadata"]["forward_step"] = max_depth - depth + 1
    for grandchild in grandchildren:
        _assign_forward_step(grandchild, max_depth, depth + 1)


def _assign_tree_depth(tree_dict: Dict[str, Any], depth: int = 1) -> int:
    """
    Assign the tree_depth property of each reaction node,
    which is the depth of the node in the tree relative to the root node.
    First reaction is at depth 1.

    :param tree_dict: the current molecule node
    :param depth: the depth of the current molecule node + 1
    :return: the maximum depth of the tree
    """
    children = tree_dict.get("children")
    if children is None:
        return depth - 1
    grandchildren = children[0]["children"]
    children[0]["metadata"]["tree_depth"] = depth
    child_depths = []
    for grandchild in grandchildren:
        child_depths.append(_assign_tree_depth(grandchild, depth + 1))
    return max(child_depths)


def _collect_atom_mapped_smiles(tree_dict: Dict[str, Any], smiles: List[str]) -> None:
    """
    Save atom-mapped SMILES from the reaction metadata in a list

    :param tree_dict: the current molecule node
    :param smiles: the list of collect reaction SMILES
    """
    children = tree_dict.get("children")
    if children is None:
        return
    smiles.append(children[0]["metadata"]["mapped_reaction_smiles"])
    for grandchild in children[0]["children"]:
        _collect_atom_mapped_smiles(grandchild, smiles)


def _collect_intermediates(tree_dict: Dict[str, Any], intermediates: Dict[str, int]) -> None:
    """
    Traverse the tree and collect SMILES of molecule nodes that has children

    :param tree_dict: the current molecule node
    :param intermediates: the list of collected leaves and their counts
    """
    if tree_dict["type"] == "mol" and tree_dict.get("children"):
        intermediates[tree_dict["smiles"]] += 1
    for child in tree_dict.get("children", []):
        _collect_intermediates(child, intermediates)


def _collect_leaves(tree_dict: Dict[str, Any], leaves: Dict[str, int]) -> None:
    """
    Traverse the tree and collect SMILES of molecule nodes that has no children

    :param tree_dict: the current molecule node
    :param leaves: the list of collected leaves and their counts
    """
    children = tree_dict.get("children", [])
    if children:
        for child in children:
            _collect_leaves(child, leaves)
    else:
        leaves[tree_dict["smiles"]] += 1


def _collect_ngrams(
    tree_dict: Dict[str, Any],
    nitems: int,
    metadata_key: str,
    result: List[Tuple[Any, ...]],
    accumulation: List[str] = None,
):
    """
    Collect ngrams from reaction metadata

    :param tree_dict: the current reaction node in the recursion
    :param nitems: the length of the gram
    :param metadata_key: the metadata to extract
    :param result: the collected ngrams
    :param accumulation: the accumulate items in the recursion
    :raise ValueError: if this routine is initialized from a molecule node
    """
    accumulation = accumulation or []
    if tree_dict["type"] == "mol":
        raise ValueError("Found _collect_ngrams at molecule node. This should not happen.")

    data = tree_dict.get("metadata", {}).get(metadata_key)
    accumulation.append(data)

    if len(accumulation) == nitems:
        result.append(tuple(accumulation))
        accumulation.pop(0)

    for mol_child in tree_dict["children"]:
        for rxn_grandchild in mol_child.get("children", []):
            _collect_ngrams(rxn_grandchild, nitems, metadata_key, result, list(accumulation))


def _collect_reaction_data(tree_dict: Dict[str, Any], data: List[Dict[str, Any]]) -> None:
    """
    Save the reaction metadata to a list and augment it
    with the un-mapped reaction SMILES and reaction depth

    :param tree_dict: the current molecule node
    :param smiles: the list of collect reaction data
    """
    children = tree_dict.get("children")
    if children is None:
        return
    grandchildren = children[0]["children"]
    reactants = join_smiles_from_reaction([grandchild["smiles"] for grandchild in grandchildren])
    metadata = deepcopy(children[0]["metadata"])
    metadata["reaction_smiles"] = f"{reactants}>>{tree_dict['smiles']}"
    data.append(metadata)
    for grandchild in grandchildren:
        _collect_reaction_data(grandchild, data)


def _collect_reaction_smiles(tree_dict: Dict[str, Any], smiles: List[str]) -> None:
    """
    Save reaction SMILES from the molecule nodes

    :param tree_dict: the current molecule node
    :param smiles: the list of collect reaction SMILES
    """
    children = tree_dict.get("children")
    if children is None:
        return
    grandchildren = children[0]["children"]
    reactants = join_smiles_from_reaction([grandchild["smiles"] for grandchild in grandchildren])
    smiles.append(f"{reactants}>>{tree_dict['smiles']}")
    for grandchild in grandchildren:
        _collect_reaction_smiles(grandchild, smiles)


def _copy_mapping_from_datamap(tree_dict: Dict[str, Any], datamap: Dict[str, Dict[str, str]]) -> None:
    """
    Store the generated atom-mapping and classification to the reaction
    metadata.

    :param tree_dict: the current molecule node
    :param datamap: a mapping from original reaction SMILES to
                    NameRxn classication and mapping as well
                    as rxnmapper mapping
    """
    children = tree_dict.get("children")
    if children is None:
        return
    grandchildren = children[0]["children"]
    reactants = join_smiles_from_reaction([grandchild["smiles"] for grandchild in grandchildren])
    rxnsmi = f"{reactants}>>{tree_dict['smiles']}"
    metadata = children[0].get("metadata", {})
    metadata["classification"] = datamap[rxnsmi]["NMC"]
    metadata["mapped_reaction_smiles"] = datamap[rxnsmi]["mapped_smiles"]
    metadata = children[0]["metadata"] = metadata
    for grandchild in grandchildren:
        _copy_mapping_from_datamap(grandchild, datamap)


def _extract_chain_molecules(
    tree_dict: Dict[str, Any],
    chain_molecules: List[Dict[str, Any]],
    step: int = None,
    parent_hash: str = None,
) -> None:
    """
    Extract molecules in the route in a format that is suitable for
    extracting the chains or sequences of the route.

    Each molecule is represented by a dictionary that contains
        * smiles - the SMILES string of the molecule
        * parent_hash - the hash of the parent molecule
        * step - the synthesis (forward) step number
        * classification - the reaction classification of the reaction
                           applied to the molecule
        * reaction_id - the reaction id of the reaction applied to the molecule
        * hash - the hash of the molecule (SMILES+step), for non-leaf molecules

    :param tree_dict: the current molecule node
    :param chain_molecules: the extracted molecules
    :param step: the forward step of the molecule
    :param parent_hash: the hash of the previous molecule node
    """
    mol_dict = {
        "smiles": tree_dict["smiles"],
        "parent_hash": parent_hash,
    }
    children = tree_dict.get("children")
    if not children:
        mol_dict["step"] = step or 0
        mol_dict["classification"] = ""
        mol_dict["reaction_id"] = None
    else:
        metadata = children[0]["metadata"]
        step = step or metadata["forward_step"]
        mol_dict["step"] = step
        mol_dict["classification"] = metadata.get("classification", "")
        mol_dict["reaction_id"] = metadata.get("id", metadata.get("ID"))
        mol_dict["hash"] = hash((mol_dict["smiles"], mol_dict["step"]))
        for grandchild in children[0]["children"]:
            _extract_chain_molecules(grandchild, chain_molecules, step - 1, mol_dict["hash"])
    chain_molecules.append(mol_dict)


def _find_leaves_not_in_stock(tree_dict: Dict[str, Any]) -> None:
    """
    Traverse the tree and check the `in_stock` value of molecule
    nodes without children (leaves). If one is found that is not
    in stock this raises an exception, which stops the traversal.
    """
    children = tree_dict.get("children", [])
    if not children and not tree_dict.get("in_stock", True):
        raise ValueError(f"child not in stock {tree_dict}")
    elif children:
        for child in children:
            _find_leaves_not_in_stock(child)


def _inherit_atom_mapping(tree_dict: Dict[str, Any], new_atomnum: int, parent_smiles: str = "") -> None:
    """
    Replace atom-mapping in the route so that it keeps the atom-mapping
    from the root node, i.e. the target molecule

    The matching of atom-mapped SMILES and unmapped SMILES are
    based on partial InChI key without stereoinformation. Very rarely
    the atom-mapping tool inverts stereochemistry, and this works
    in those situations.

    :param tree_dict: the current molecule node
    :param new_atomnum: the atom mapping number to assign to unmapped atoms
    :param parent_smiles: the SMILE of the parent molecule node
    """
    # If at leaf node, just return
    if not tree_dict.get("children", []):
        return

    # Extract reaction and atom-mapped reaction SMILES
    reaction_dict = tree_dict["children"][0]
    mapped_rsmi0 = reaction_dict["metadata"]["mapped_reaction_smiles"]

    # If we are at the root, i.e. target node, we just use the atom-mapping as-is
    if parent_smiles == "":
        mapped_rsmi = mapped_rsmi0
    else:
        _, _, product0 = split_rsmi(mapped_rsmi0)
        remapping = _find_remapping(parent_smiles, product0)
        product_mol = Chem.MolFromSmiles(product0)
        for atom in product_mol.GetAtoms():
            if atom.GetAtomMapNum() and atom.GetAtomMapNum() not in remapping:
                new_atomnum += 1
                remapping[atom.GetAtomMapNum()] = new_atomnum
        mapped_rsmi = _apply_remapping(mapped_rsmi0, remapping, keep_original=False)

    reaction_dict["metadata"]["mapped_reaction_smiles"] = mapped_rsmi

    # Recursing to reactants
    inchi2mapped = {
        smiles2inchikey(smiles, ignore_stereo=True): smiles
        for smiles in split_smiles_from_reaction(split_rsmi(mapped_rsmi)[0])
    }
    for grandchild in reaction_dict["children"]:
        # When a molecule is a complex use each part to inherit atom mapping
        for smiles in split_smiles_from_reaction(grandchild["smiles"]):
            inchi = smiles2inchikey(smiles, ignore_stereo=True)
            _inherit_atom_mapping(grandchild, new_atomnum, inchi2mapped[inchi])


def _remap_reactions(tree_dict: Dict[str, Any], remapping: Dict[int, int]) -> None:
    """
    Remapp each of the reaction using a remapping dictionary

    :param tree_dict: the current molecule node
    :param remapping: the remapping dictionary
    """
    children = tree_dict.get("children")
    if children is None:
        return
    reaction_smiles = children[0]["metadata"]["mapped_reaction_smiles"]
    children[0]["metadata"]["mapped_reaction_smiles"] = _apply_remapping(reaction_smiles, remapping, keep_original=True)
    for grandchild in children[0]["children"]:
        _remap_reactions(grandchild, remapping)



"""
Module containing helper classes to compute the distance between to reaction trees using the APTED method
Since APTED is based on ordered trees and the reaction trees are unordered, plenty of
heuristics are implemented to deal with this.
"""

from __future__ import annotations

import itertools
import math
from copy import deepcopy
from logging import getLogger
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
from apted import APTED as Apted

from rxnutils.routes.base import SynthesisRoute
from rxnutils.routes.ted.utils import AptedConfig, StandardFingerprintFactory, TreeContent

StrDict = Dict[str, Any]
_FloatIterator = Iterable[float]


class ReactionTreeWrapper:
    """
    Wrapper for a reaction tree that can calculate distances between
    trees.

    :param route: the synthesis route to wrap
    :param content: the content of the route to consider in the distance calculation, default 'molecules'
    :param exhaustive_limit: if the number of possible ordered trees are below this limit create them all, default 20
    :param fp_factory: the factory of the fingerprint, Morgan fingerprint for molecules and reactions by default
    :param dist_func: the distance function to use when renaming nodes
    """

    _index_permutations = {n: list(itertools.permutations(range(n), n)) for n in range(1, 8)}

    def __init__(
        self,
        route: SynthesisRoute,
        content: Union[str, TreeContent] = TreeContent.MOLECULES,
        exhaustive_limit: int = 20,
        fp_factory: Callable[[StrDict, Optional[StrDict]], None] = None,
        dist_func: Callable[[np.ndarray, np.ndarray], float] = None,
    ) -> None:
        single_node_tree = not bool(route.reaction_smiles())
        if single_node_tree and content == TreeContent.REACTIONS:
            raise ValueError("Cannot create wrapping with content = reactions for a tree without reactions")

        self._logger = getLogger()
        # Will convert string input automatically
        self._content = TreeContent(content)
        self._base_tree = deepcopy(route.reaction_tree)

        self._fp_factory = fp_factory or StandardFingerprintFactory()
        self._add_fingerprints(self._base_tree)

        if self._content != TreeContent.MOLECULES and not single_node_tree:
            self._add_fingerprints(self._base_tree["children"][0], self._base_tree)

        if self._content == TreeContent.MOLECULES:
            self._base_tree = self._remove_children_nodes(self._base_tree)
        elif not single_node_tree and self._content == TreeContent.REACTIONS:
            self._base_tree = self._remove_children_nodes(self._base_tree["children"][0])

        self._trees = []
        self._tree_count, self._node_index_list = self._inspect_tree()
        self._enumeration = self._tree_count <= exhaustive_limit

        if self._enumeration:
            self._create_all_trees()
        else:
            self._trees.append(self._base_tree)

        self._dist_func = dist_func

    @property
    def info(self) -> StrDict:
        """Return a dictionary with internal information about the wrapper"""
        return {
            "content": self._content,
            "tree count": self._tree_count,
            "enumeration": self._enumeration,
        }

    @property
    def first_tree(self) -> StrDict:
        """Return the first created ordered tree"""
        return self._trees[0]

    @property
    def trees(self) -> List[StrDict]:
        """Return a list of all created ordered trees"""
        return self._trees

    def distance_iter(self, other: "ReactionTreeWrapper", exhaustive_limit: int = 20) -> _FloatIterator:
        """
        Iterate over all distances computed between this and another tree

        There are three possible enumeration of distances possible dependent
        on the number of possible ordered trees for the two routes that are compared

        * If the product of the number of possible ordered trees for both routes are
          below `exhaustive_limit` compute the distance between all pair of trees
        * If both self and other has been fully enumerated (i.e. all ordered trees has been created)
          compute the distances between all trees of the route with the most ordered trees and
          the first tree of the other route
        * Compute `exhaustive_limit` number of distances by shuffling the child order for
          each of the routes.

        The rules are applied top-to-bottom.

        :param other: another tree to calculate distance to
        :param exhaustive_limit: used to determine what type of enumeration to do
        :yield: the next computed distance between self and other
        """
        if self._tree_count * other.info["tree count"] < exhaustive_limit:
            yield from self._distance_iter_exhaustive(other)
        elif self._enumeration or other.info["enumeration"]:
            yield from self._distance_iter_semi_exhaustive(other)
        else:
            yield from self._distance_iter_random(other, exhaustive_limit)

    def distance_to(self, other: "ReactionTreeWrapper", exhaustive_limit: int = 20) -> float:
        """
        Calculate the minimum distance from this route to another route

        Enumerate the distances using `distance_iter`.

        :param other: another tree to calculate distance to
        :param exhaustive_limit: used to determine what type of enumeration to do
        :return: the minimum distance
        """
        min_dist = 1e6
        min_iter = -1
        for iteration, distance in enumerate(self.distance_iter(other, exhaustive_limit)):
            if distance < min_dist:
                min_iter = iteration
                min_dist = distance
        self._logger.debug(f"Found minimum after {min_iter} iterations")
        return min_dist

    def distance_to_with_sorting(self, other: "ReactionTreeWrapper") -> float:
        """
        Compute the distance to another tree, by simpling sorting the children
        of both trees. This is not guaranteed to return the minimum distance.

        :param other: another tree to calculate distance to
        :return: the distance
        """
        config = AptedConfig(sort_children=True, dist_func=self._dist_func)
        return Apted(self.first_tree, other.first_tree, config).compute_edit_distance()

    def _add_fingerprints(self, tree: StrDict, parent: StrDict = None) -> None:
        if "fingerprint" not in tree:
            try:
                self._fp_factory(tree, parent)
            except ValueError:
                pass
        if "fingerprint" not in tree:
            tree["fingerprint"] = []
        tree["sort_key"] = "".join(f"{digit}" for digit in tree["fingerprint"])
        if "children" not in tree:
            tree["children"] = []

        for child in tree["children"]:
            for grandchild in child["children"]:
                self._add_fingerprints(grandchild, child)

    def _create_all_trees(self) -> None:
        self._trees = []
        # Iterate over all possible combinations of child order
        for order_list in itertools.product(*self._node_index_list):
            self._trees.append(self._create_tree_recursively(self._base_tree, list(order_list)))

    def _create_tree_recursively(
        self,
        node: StrDict,
        order_list: List[List[int]],
    ) -> StrDict:
        new_tree = self._make_base_copy(node)
        children = node.get("children", [])
        if children:
            child_order = order_list.pop(0)
            assert len(child_order) == len(children)
            new_children = [self._create_tree_recursively(child, order_list) for child in children]
            new_tree["children"] = [new_children[idx] for idx in child_order]
        return new_tree

    def _distance_iter_exhaustive(self, other: "ReactionTreeWrapper") -> _FloatIterator:
        self._logger.debug(f"APTED: Exhaustive search. {len(self.trees)} {len(other.trees)}")
        config = AptedConfig(randomize=False, dist_func=self._dist_func)
        for tree1, tree2 in itertools.product(self.trees, other.trees):
            yield Apted(tree1, tree2, config).compute_edit_distance()

    def _distance_iter_random(self, other: "ReactionTreeWrapper", ntimes: int) -> _FloatIterator:
        self._logger.debug(f"APTED: Heuristic search. {len(self.trees)} {len(other.trees)}")
        config = AptedConfig(randomize=False, dist_func=self._dist_func)
        yield Apted(self.first_tree, other.first_tree, config).compute_edit_distance()

        config = AptedConfig(randomize=True, dist_func=self._dist_func)
        for _ in range(ntimes):
            yield Apted(self.first_tree, other.first_tree, config).compute_edit_distance()

    def _distance_iter_semi_exhaustive(self, other: "ReactionTreeWrapper") -> _FloatIterator:
        self._logger.debug(f"APTED: Semi-exhaustive search. {len(self.trees)} {len(other.trees)}")
        if len(self.trees) < len(other.trees):
            first_wrapper = self
            second_wrapper = other
        else:
            first_wrapper = other
            second_wrapper = self

        config = AptedConfig(randomize=False, dist_func=self._dist_func)
        for tree1 in first_wrapper.trees:
            yield Apted(tree1, second_wrapper.first_tree, config).compute_edit_distance()

    def _inspect_tree(self) -> Tuple[int, List[List[int]]]:
        """
        Find the number of children for each node in the tree, which
        will be used to compute the number of possible combinations of child orders

        Also accumulate the possible child orders for the nodes.
        """

        def _recurse_tree(node):
            children = node.get("children", [])
            nchildren = len(children)
            permutations.append(math.factorial(nchildren))

            if nchildren > 0:
                node_index_list.append(list(self._index_permutations[nchildren]))
            for child in children:
                _recurse_tree(child)

        permutations: List[int] = []
        node_index_list: List[List[int]] = []
        _recurse_tree(self._base_tree)
        if not permutations:
            return 0, []
        return int(np.prod(permutations)), node_index_list

    @staticmethod
    def _make_base_copy(node: StrDict) -> StrDict:
        return {
            "type": node["type"],
            "smiles": node.get("smiles", ""),
            "metadata": node.get("metadata"),
            "fingerprint": node["fingerprint"],
            "sort_key": node["sort_key"],
            "children": [],
        }

    @staticmethod
    def _remove_children_nodes(tree: StrDict) -> StrDict:
        new_tree = ReactionTreeWrapper._make_base_copy(tree)

        if tree.get("children"):
            new_tree["children"] = []
            for child in tree["children"]:
                new_tree["children"].extend(
                    [ReactionTreeWrapper._remove_children_nodes(grandchild) for grandchild in child.get("children", [])]
                )
        return new_tree



""" Routines for augmenting chemical reactions
"""

from rxnutils.chem.utils import split_rsmi

_SINGLE_REACTANT_REAGENTS = {"10.1.1": "Br", "10.1.2": "Cl"}


def single_reactant_augmentation(smiles: str, classification: str) -> str:
    """
    Augment single-reactant reaction with additional reagent if possible
    based on the classification of the reaction
    :param smiles: the reaction SMILES to augment
    :param classification: the classification of the reaction or an empty string
    :return: the processed SMILES
    """
    reactants = split_rsmi(smiles)[0]
    if "." in reactants:
        return smiles
    classification = classification.split(" ")[0]
    new_reactant = _SINGLE_REACTANT_REAGENTS.get(classification)
    if new_reactant:
        return new_reactant + "." + smiles
    return smiles



""" Module containing routes to validate AiZynthFinder-like input dictionaries """

from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, StringConstraints, ValidationError, conlist
from typing_extensions import Annotated

StrDict = Dict[str, Any]


class ReactionNode(BaseModel):
    """Node representing a reaction"""

    type: Annotated[str, StringConstraints(pattern=r"^reaction$")]
    children: List[MoleculeNode]


class MoleculeNode(BaseModel):
    """Node representing a molecule"""

    smiles: str
    type: Annotated[str, StringConstraints(pattern=r"^mol$")]
    children: Optional[conlist(ReactionNode, min_length=1, max_length=1)] = None


MoleculeNode.update_forward_refs()


def validate_dict(dict_: StrDict) -> None:
    """
    Check that the route dictionary is a valid structure

    :param dict_: the route as dictionary
    """
    try:
        MoleculeNode(**dict_, extra="ignore")
    except ValidationError as err:
        raise ValueError(f"Invalid input: {err.json()}")



"""Module containing template validation actions"""

from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar, Sequence, Set

import pandas as pd
from rdkit import Chem, RDLogger
from rdkit.Chem import AllChem

from rxnutils.chem.template import ReactionTemplate
from rxnutils.chem.utils import split_rsmi, split_smiles_from_reaction
from rxnutils.pipeline.base import action, global_apply

rd_logger = RDLogger.logger()
rd_logger.setLevel(RDLogger.CRITICAL)


@action
@dataclass
class CountTemplateComponents:
    """Action for counting template components"""

    pretty_name: ClassVar[str] = "count_template_components"
    in_column: str
    nreactants_column: str = "nreactants"
    nreagents_column: str = "nreagents"
    nproducts_column: str = "nproducts"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        new_data = global_apply(data, self._apply_row, axis=1)
        return data.assign(**{column: new_data[column] for column in new_data.columns})

    def _apply_row(self, row: pd.Series) -> pd.Series:
        if not isinstance(row[self.in_column], str):
            return pd.Series(
                {
                    self.nreactants_column: None,
                    self.nreagents_column: None,
                    self.nproducts_column: None,
                }
            )

        rd_reaction = AllChem.ReactionFromSmarts(row[self.in_column])
        return pd.Series(
            {
                self.nproducts_column: rd_reaction.GetNumReactantTemplates(),
                self.nreagents_column: rd_reaction.GetNumAgentTemplates(),
                self.nreactants_column: rd_reaction.GetNumProductTemplates(),
            }
        )

    def __str__(self) -> str:
        return f"{self.pretty_name} (counting reactants, reagents, products in template)"


@action
@dataclass
class RetroTemplateReproduction:
    """Action for checking template reproduction"""

    pretty_name: ClassVar[str] = "retro_template_reproduction"
    template_column: str
    smiles_column: str
    expected_reactants_column: str = "TemplateGivesTrueReactants"
    other_reactants_column: str = "TemplateGivesOtherReactants"
    noutcomes_column: str = "TemplateGivesNOutcomes"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        new_data = global_apply(data, self._apply_row, axis=1)
        return data.assign(**{column: new_data[column] for column in new_data.columns})

    def _apply_row(self, row: pd.Series) -> pd.Series:
        no_oututcome_response = pd.Series(
            {
                self.noutcomes_column: 0,
                self.expected_reactants_column: False,
                self.other_reactants_column: False,
            }
        )
        if not isinstance(row[self.template_column], str):
            return no_oututcome_response

        reactants, _, products = split_rsmi(row[self.smiles_column])
        # Flattening the list of SMILES, i.e. intramolecular -> intermolecular,
        # because all produced templates will be intermolecular
        expected_reactants = self._smiles_to_inchiset(
            [smi for reactant in split_smiles_from_reaction(reactants) for smi in reactant.split(".")]
        )
        template = ReactionTemplate(row[self.template_column], direction="retro")

        try:
            outcome = template.apply(products)
        except Exception:  # pylint: disable=broad-except # noqa
            return no_oututcome_response
        if not outcome:
            return no_oututcome_response

        found_expected_reactants = False
        found_other_reactants = False
        for outcome_item in outcome:
            if self._smiles_to_inchiset(outcome_item).issubset(expected_reactants):
                found_expected_reactants = True
            else:
                found_other_reactants = True
        return pd.Series(
            {
                self.noutcomes_column: len(outcome),
                self.expected_reactants_column: found_expected_reactants,
                self.other_reactants_column: found_other_reactants,
            }
        )

    @staticmethod
    def _smiles_to_inchiset(smiles_list: Sequence[str]) -> Set[str]:
        inchi_list = []
        for smi in smiles_list:
            try:
                mol = Chem.MolFromSmiles(smi)
            except Exception:  # pylint: disable=broad-except # noqa
                pass
            else:
                if mol:
                    inchi_list.append(Chem.MolToInchiKey(mol))
        return set(inchi_list)

    def __str__(self) -> str:
        return f"{self.pretty_name} (checking that retro template produce the expected reactants)"



"""
Module containing pipeline for mapping with rxnmapper
This needs to be run in an environment with rxnmapper installed
"""

from pathlib import Path

from metaflow import Parameter, step

from rxnutils.data.base_pipeline import DataBaseFlow
from rxnutils.data.mapping import main as map_data


class RxnMappingFlow(DataBaseFlow):
    """Pipeline for atom-map USPTO or ORD data with rxnmapper"""

    data_prefix = Parameter("data-prefix")

    @step
    def start(self):
        """Setup batches for mapping"""
        # pylint: disable=attribute-defined-outside-init
        self.partitions = self._create_batches(
            Path(self.folder) / f"{self.data_prefix}_data_cleaned.csv",
            Path(self.folder) / f"{self.data_prefix}_data_mapped.csv",
        )
        self.next(self.do_mapping, foreach="partitions")

    @step
    def do_mapping(self):
        """Perform atom-mapping of reactions"""
        idx, start, end = self.input
        output_filename = Path(self.folder) / f"{self.data_prefix}_data_mapped.csv.{idx}"
        if idx > -1 and not output_filename.exists():
            map_data(
                [
                    "--input",
                    str(Path(self.folder) / f"{self.data_prefix}_data_cleaned.csv"),
                    "--output",
                    str(output_filename),
                    "--batch",
                    str(start),
                    str(end),
                    "--mapper_batch_size",
                    "2",
                ]
            )
        self.next(self.join_mapping)

    @step
    def join_mapping(self, _):
        """Join batches from mapping"""
        self._combine_batches(Path(self.folder) / f"{self.data_prefix}_data_mapped.csv")
        self.next(self.end)

    @step
    def end(self):
        """Final step, just print information"""
        filename = Path(self.folder) / f"{self.data_prefix}_data_mapped.csv"
        print(f"Processed file is locate here: {filename}")


if __name__ == "__main__":
    RxnMappingFlow()



"""Routines for reading routes from various formats"""

import copy
from typing import Any, Dict, List, Sequence

import pandas as pd
from rdkit import Chem
from rdkit.Chem import AllChem

from rxnutils.chem.utils import join_smiles_from_reaction, split_rsmi, split_smiles_from_reaction
from rxnutils.routes.base import SynthesisRoute, smiles2inchikey


def read_reaction_lists(filename: str) -> List[SynthesisRoute]:
    """
    Read one or more simple lists of reactions into one or more
    retrosynthesis trees.

    Each list of reactions should be separated by an empty line.
    Each row of each reaction should contain the reaction SMILES (reactants>>products)
    and nothing else.

    Example:
    A.B>>C
    D.E>>B

    A.X>>Y
    Z>>X

    defines two retrosynthesis trees, and the first being
         A
    C ->      D
         B ->
              E

    :params filename: the path to the file with the reactions
    :returns: the list of the created trees
    """
    with open(filename, "r") as fileobj:
        all_lines = fileobj.read()

    reaction_lists = [lines.splitlines() for lines in all_lines.split("\n\n")]
    return [reactions2route(reactions) for reactions in reaction_lists]


def read_aizynthcli_dataframe(data: pd.DataFrame) -> pd.Series:
    """
    Read routes as produced by the `aizynthcli` tool of the `AiZynthFinder` package.

    :param data: the dataframe as output by `aizynthcli`
    :return: the created routes
    """

    def read_row(row: pd.Series) -> List[SynthesisRoute]:
        return [read_aizynthfinder_dict(tree) for tree in row.trees]

    return data.apply(read_row, axis=1)


def read_aizynthfinder_dict(tree: Dict[str, Any]) -> SynthesisRoute:
    """
    Read a single aizynthfinder dictionary

    :param tree: the aizynthfinder structure
    :return: the created routes
    """
    dict_ = copy.deepcopy(tree)
    _transform_retrosynthesis_atom_mapping(dict_)
    return SynthesisRoute(dict_)


def read_reactions_dataframe(
    data: pd.DataFrame,
    smiles_column: str,
    group_by: List[str],
    metadata_columns: List[str] = None,
) -> pd.Series:
    """
    Read routes from reactions stored in a pandas dataframe. The different
    routes are groupable by one or more column. Additional metadata columns
    can be extracted from the dataframe as well.

    The dataframe is grouped by the columns specified by `group_by` and
    then one routes is extracted from each subset dataframe. The function
    returns a series with the routes, which is indexable by the columns
    in the `group_by` list.

    :param data: the dataframe with reaction data
    :param smiles_column: the column with the reaction SMILES
    :param group_by: the columns that uniquely identifies each route
    :param metadata_column: additional columns to be added as metadata to each route
    :return: the created series with route.
    """

    def reaction_dataframe2routes(subdata):
        smiles = subdata[smiles_column]
        metadata = subdata[metadata_columns].to_dict("records")
        try:
            route = reactions2route(smiles, metadata)
        except ValueError:
            id_str = ", ".join([str(subdata.iloc[0][col]) for col in group_by])
            print(f"Failed to re-create synthesis tree for: ({id_str})")
            route = None
        dict_ = {col: subdata[col].iloc[0] for col in group_by}
        dict_["route"] = route
        return route

    metadata_columns = metadata_columns or []
    grouped = data.groupby(group_by)[data.columns.to_list()]
    return grouped.apply(reaction_dataframe2routes)


def reactions2route(reactions: Sequence[str], metadata: Sequence[Dict[str, Any]] = None) -> SynthesisRoute:
    """
    Convert a list of reactions into a retrosynthesis tree

    This is based on matching partial InChI keys of the reactants in one
    reaction with the partial InChI key of a product.

    :params reactions: list of reaction SMILES
    :returns: the created trees
    """

    def make_dict(product_smiles):
        dict_ = {
            "type": "mol",
            "smiles": product_smiles,
        }
        product_inchi = smiles2inchikey(product_smiles, ignore_stereo=True)
        reaction = product2reaction.get(product_inchi)
        if reaction is not None:
            metadata = dict(reaction["metadata"])
            metadata["reaction_smiles"] = join_smiles_from_reaction(reaction["reactants"]) + ">>" + reaction["product"]
            dict_["children"] = [
                {
                    "type": "reaction",
                    "metadata": metadata,
                    "children": [make_dict(smiles) for smiles in reaction["reactants"]],
                }
            ]
        return dict_

    metadata = metadata or [dict() for _ in range(len(reactions))]
    all_reactants = set()
    product2reaction = {}
    inchi_map = {}
    for reaction, meta in zip(reactions, metadata):
        reactants_smiles, _, product_smiles = split_rsmi(reaction)
        product_smiles = Chem.CanonSmiles(product_smiles)
        partial_product_inchi = smiles2inchikey(product_smiles, ignore_stereo=True)
        inchi_map[partial_product_inchi] = product_smiles
        reactants = split_smiles_from_reaction(reactants_smiles)
        try:
            reactants = [Chem.CanonSmiles(smi) for smi in reactants]
        except:
            raise ValueError(f"Cannot canonicalize SMILES: {reactants_smiles}")
        all_reactants = all_reactants.union([smiles2inchikey(smi, ignore_stereo=True) for smi in reactants])
        product2reaction[partial_product_inchi] = {
            "smiles": reaction,
            "product": product_smiles,
            "reactants": reactants,
            "metadata": meta,
        }

    only_products = set(product2reaction.keys()) - all_reactants
    if len(only_products) != 1:
        raise ValueError(f"Could not identify one and only one target product: {only_products}")

    target_inchi = list(only_products)[0]
    return SynthesisRoute(make_dict(inchi_map[target_inchi]))


def read_rdf_file(filename: str) -> SynthesisRoute:
    def finish_reaction():
        if not rxnblock:
            return
        rxn = AllChem.ReactionFromRxnBlock("\n".join(rxnblock), sanitize=False, strictParsing=False)
        reactions.append(AllChem.ReactionToSmiles(rxn))

    with open(filename, "r") as fileobj:
        lines = fileobj.read().splitlines()

    reactions = []
    rxnblock = []
    read_rxn = skip_entry = False
    for line in lines:
        if line.startswith("$RFMT"):
            read_rxn = skip_entry = False
            finish_reaction()
            rxnblock = []
        elif line.startswith(("$MFMT", "$DATUM")):
            # Ignore MFMT and DATUM entries for now
            skip_entry = True
        elif skip_entry:
            continue
        elif line.startswith("$RXN"):
            rxnblock = [line]
            read_rxn = True
        elif read_rxn:
            rxnblock.append(line)
    finish_reaction()

    return reactions2route(reactions)


def _transform_retrosynthesis_atom_mapping(tree_dict: Dict[str, Any]) -> None:
    """
    Routes output from AiZynth has atom-mapping from the template-based model,
    but it needs to be processed
    1. Remove atom-mapping from reactants not in product
    2. Reverse reaction SMILES
    """
    if not tree_dict.get("children", []):
        return

    reaction_dict = tree_dict["children"][0]
    mapped_rsmi = reaction_dict["metadata"]["mapped_reaction_smiles"]

    product, _, reactants = split_rsmi(mapped_rsmi)
    product_mol = Chem.MolFromSmiles(product)
    product_maps = {atom.GetAtomMapNum() for atom in product_mol.GetAtoms() if atom.GetAtomMapNum()}
    reactant_mols = []
    for smiles in reactants.split("."):
        mol = Chem.MolFromSmiles(smiles)
        for atom in mol.GetAtoms():
            if atom.GetAtomMapNum() and atom.GetAtomMapNum() not in product_maps:
                atom.SetAtomMapNum(0)
        reactant_mols.append(mol)

    mapped_rsmi = ">".join(
        [
            join_smiles_from_reaction([Chem.MolToSmiles(mol) for mol in reactant_mols]),
            "",
            Chem.MolToSmiles(product_mol),
        ]
    )
    reaction_dict["metadata"]["mapped_reaction_smiles"] = mapped_rsmi

    for grandchild in reaction_dict["children"]:
        _transform_retrosynthesis_atom_mapping(grandchild)



import os
from pathlib import Path
from typing import Any, List, Optional, Tuple

import numpy as np
import pandas as pd
from scipy import sparse


def nlines(filename: str) -> int:
    """Count and return the number of lines in a file"""
    with open(filename, "rb") as fileobj:
        return sum(1 for line in fileobj)


def combine_batches(
    filename: str,
    nbatches: int,
    read_func: Any,
    write_func: Any,
    combine_func: Any,
) -> None:
    if Path(filename).exists():
        return

    data = None
    for idx in range(nbatches):
        temp_data, filename2 = read_func(filename, idx)
        if data is None:
            data = temp_data
        else:
            data = combine_func(data, temp_data)
        os.remove(filename2)
    write_func(data, filename)


def combine_csv_batches(filename: str, nbatches: int) -> None:
    """
    Combine CSV batches to one master file

    The batch files are removed from disc

    :param filename: the filename of the master file
    :param nbatches: the number of batches
    """

    def _combine_csv(data: pd.DataFrame, temp_data: pd.DataFrame) -> pd.DataFrame:
        return pd.concat([data, temp_data])

    def _read_csv(filename: str, idx: int) -> pd.DataFrame:
        filename2 = f"{filename}.{idx}"
        return pd.read_csv(filename2, sep="\t"), filename2

    def _write_csv(data: pd.DataFrame, filename: str) -> None:
        data.to_csv(filename, index=False, sep="\t")

    combine_batches(filename, nbatches, _read_csv, _write_csv, _combine_csv)


def combine_numpy_array_batches(filename: str, nbatches: int) -> None:
    """
    Combine numpy array batches to one master file
    The batch files are removed from disc
    :param filename: the filename of the master file
    :param nbatches: the number of batches
    """

    def _read_array(filename: str, idx: int) -> Any:
        filename2 = filename.replace(".npz", f".{idx}.npz")
        return np.load(filename2)["arr_0"], filename2

    def _write_array(data: np.ndarray, filename: str) -> None:
        np.savez(filename, data, compressed=True)

    def _combine_array(data: np.ndarray, temp_data: np.ndarray) -> np.ndarray:
        return np.hstack([data, temp_data])

    return combine_batches(filename, nbatches, _read_array, _write_array, _combine_array)


def combine_sparse_matrix_batches(filename: str, nbatches: int) -> None:
    """
    Combine sparse matrix batches to one master file

    The batch files are removed from disc

    :param filename: the filename of the master file
    :param nbatches: the number of batches
    """

    def _read_matrix(filename: str, idx: int) -> Any:
        filename2 = filename.replace(".npz", f".{idx}.npz")
        return sparse.load_npz(filename2), filename2

    def _write_matrix(data: Any, filename: str) -> None:
        sparse.save_npz(filename, data, compressed=True)

    def _combine_matrix(data: pd.DataFrame, temp_data: pd.DataFrame) -> pd.DataFrame:
        return sparse.vstack([data, temp_data])

    combine_batches(filename, nbatches, _read_matrix, _write_matrix, _combine_matrix)


def create_csv_batches(
    filename: str, nbatches: int, output_filename: Optional[str] = None
) -> List[Tuple[int, int, int]]:
    """
    Create batches for reading a splitted CSV-file

    The batches will be in  the form of a tuple with three indices:
        * Batch index
        * Start index
        * End index

    :param filename: the CSV file to make batches of
    :param nbatches: the number of batches
    :param output_filename:
    :return: the created batches
    """
    if output_filename and Path(output_filename).exists():
        return [(-1, None, None)]

    file_size = nlines(filename) - 1  # Header should not be counted for batch size calculations
    nbatches = min(file_size, nbatches)  # Adjust the number of batches to the size of the file
    batch_size, remainder = divmod(file_size, nbatches)
    stop = 1  # 1-indexed to account for header in the .csv file
    batches = []
    for partition_idx in range(1, nbatches + 1):
        start = stop
        stop += batch_size + 1 if partition_idx <= remainder else batch_size
        batches.append((partition_idx - 1, start, stop))
    return batches


def read_csv_batch(filename: str, batch: Tuple[int, ...] = None, **kwargs: Any) -> pd.DataFrame:
    """
    Read parts of a CSV file as specified by a batch

    :param filename: the path to the CSV file on disc
    :param batch: the batch specification as returned by `create_csv_batches`
    """
    if batch is None:
        return pd.read_csv(filename, **kwargs)
    if len(batch) == 3:
        _, batch_start, batch_end = batch
    elif len(batch) == 2:
        batch_start, batch_end = batch
    else:
        raise ValueError(f"The batch specification can only be 2 or 3 not {len(batch)}")
    return pd.read_csv(
        filename,
        nrows=batch_end - batch_start,
        skiprows=range(1, batch_start),
        **kwargs,
    )



"""Module containing an API to the Reaction InChI program"""

import logging
import os
import subprocess
import sys
import tempfile
from collections import namedtuple

from rdkit.Chem import AllChem

from rxnutils.chem.rinchi import download_rinchi
from rxnutils.chem.rinchi.download_rinchi import PLATFORM2FOLDER, RInChIError

RInChIStructure = namedtuple("RInChI", "rinchi rauxinfo long_rinchikey short_rinchikey web_rinchikey")


def generate_rinchi(reaction_smiles: str) -> RInChIStructure:
    """Generate RInChI from Reaction SMILES.

    :param reaction_smiles: Reaction SMILES
    :raises RInChIError: When there is an error with RInChI generation.
    :return: Namedtuple with the generated RInChI.
    """
    if sys.platform not in PLATFORM2FOLDER:
        raise RInChIError("RInChI software not supported on this platform")

    reaction = AllChem.ReactionFromSmarts(reaction_smiles)
    reaction.Initialize()
    nwarn, nerror, nreactants, nproducts, labels = AllChem.PreprocessReaction(reaction)
    logging.debug(f"Number of warnings: {nwarn}")
    logging.debug(f"Number of preprocessing errors: {nerror}")
    logging.debug(f"Number of reactants in reaction: {nreactants}")
    logging.debug(f"Number of products in reaction: {nproducts}")
    logging.debug(f"Preprocess labels added: {labels}")

    rxn_block = AllChem.ReactionToRxnBlock(reaction)

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as fileobj:
        tmp_file = fileobj.name
        fileobj.write(rxn_block)
    logging.debug(f"Temp RXN file: {tmp_file}")

    try:
        rinchi_data = _rinchi_cli(rxn_file=tmp_file)
        logging.debug(rinchi_data)
    except RInChIError:  # pylint: disable=try-except-raise
        raise
    finally:
        os.unlink(tmp_file)

    return rinchi_data


def _rinchi_cli(rxn_file: str, options: str = None) -> RInChIStructure:
    """Run the commandline to generate RInChI.

    :param rxn_file: Path for RXN file
    :param options: Additional options from rinchi_cmdline, defaults to None
    :raises RInChIError: When there is an error with RInChI generation.
    :return: Namedtuple with the generated RInChI.
    """
    rinchi_path = download_rinchi.main()
    cmd = [os.path.join(rinchi_path, "rinchi_cmdline"), rxn_file]
    if options:
        cmd += [options]

    # Note: rinchi_cmdline doesn't like fuzzyness...
    # I.e.: Neighbouring bonds to double bond, expected to be defined...
    # pylint: disable=subprocess-run-check
    process = subprocess.run(
        cmd,
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    if process.returncode:
        raise RInChIError(process.stderr)
    lines = process.stdout.strip().split("\n")
    rinchi_data = RInChIStructure(
        rinchi=lines[0],
        rauxinfo=lines[1],
        long_rinchikey=lines[2],
        short_rinchikey=lines[3],
        web_rinchikey=lines[4],
    )

    return rinchi_data


if __name__ == "__main__":
    print(generate_rinchi(sys.argv[1]))



from __future__ import annotations

import argparse
from collections import OrderedDict
from typing import List, Optional, Sequence

import pandas as pd
from rdkit import Chem

from rxnutils.chem.utils import split_rsmi


def _get_atom_identifier(atom: Chem.rdchem.Atom) -> str:
    """
    Get atom identifier for neighborhood identification.
    The identifier is either the atom-map number if available, otherwise the symbol.
    :param atom: rdkit atom
    :return: an atom identifier string
    """
    atom_id = atom.GetAtomMapNum()
    if atom_id == 0:
        atom_id = atom.GetSymbol()
    return str(atom_id)


def _get_bond_environment_identifier(atoms: Sequence[Chem.rdchem.Atom], bond: Chem.rdchem.Bond) -> str:
    """
    Get the environment of a specific bond.

    :param atoms: atoms in the molecule.
    :param bond: bond for which the environment should be specified
    :return: string representation of the bond environment
    """
    atom_map1 = _get_atom_identifier(atoms[bond.GetBeginAtomIdx()])
    atom_map2 = _get_atom_identifier(atoms[bond.GetEndAtomIdx()])
    bond_order = bond.GetBondType()
    atom_map1, atom_map2 = sorted([atom_map1, atom_map2])
    return f"{atom_map1}_{atom_map2}_{bond_order}"


def _get_atomic_neighborhoods(smiles: str) -> OrderedDict[int, List[str]]:
    """
    Obtains a dictionary containing each atom (atomIdx) and a list of its
    bonding environment.

    :param smiles: Atom-mapped SMILES string
    :return: A dictionary containing each atom (atomIdx) and a list of its
        bonding environment identifiers.
    """

    mol = Chem.MolFromSmiles(smiles)
    atoms = mol.GetAtoms()

    neighbor_dict = {}
    for atom in atoms:
        bonds_list = []
        if atom.GetAtomMapNum() != 0:
            for bond in atom.GetBonds():

                bonds_list.append(_get_bond_environment_identifier(atoms, bond))

            neighbor_dict[atom.GetAtomMapNum()] = sorted(bonds_list)
    ordered_neighbor_dict = OrderedDict(sorted(neighbor_dict.items()))

    return ordered_neighbor_dict


def get_atom_list(reactants_smiles: str, product_smiles: str) -> List[int]:
    """
    Given two sets of SMILES strings corresponding to a set of reactants and products,
    returns a list of atomIdxs for which the atomic environment has changed,
    as defined by a change in the bonds.

    :param reactants_smiles: Atom-mapped SMILES string for the reactant(s)
    :param product_smiles: Atom-mapped SMILES string for the product(s)
    :return: List of atoms (atomIdx) for which the atomic environment has changed
    """

    ordered_reactant_neighbor_dict = _get_atomic_neighborhoods(reactants_smiles)
    ordered_product_neighbor_dict = _get_atomic_neighborhoods(product_smiles)

    all_indices = set(ordered_product_neighbor_dict.keys()) | set(ordered_reactant_neighbor_dict.keys())

    # Checks to see equivlence of atomic enviroments.
    # If environment changed, then add atom to list
    atom_list = [
        atom_map
        for atom_map in all_indices
        if ordered_reactant_neighbor_dict.get(atom_map, []) != ordered_product_neighbor_dict.get(atom_map, [])
    ]

    return atom_list


def atom_map_tag_reactants(mapped_rxn: str) -> str:
    """
    Given atom-mapped reaction, returns disconnection site-tagged reactants where atoms
    with changed atom environment are represented by [<atom>:1].

    :param mapped_rxn: Atom-mapped reaction SMILES
    :return: SMILES of the reactants containing tags corresponding to atoms changed in the
        reaction.
    """
    reactants_smiles, _, product_smiles = split_rsmi(mapped_rxn)

    reactants_mol = Chem.MolFromSmiles(reactants_smiles)
    atom_list = get_atom_list(reactants_smiles, product_smiles)

    # Set atoms in product with a different environment in reactants to 1
    for atom in reactants_mol.GetAtoms():
        if atom.GetAtomMapNum() in atom_list:
            atom.SetAtomMapNum(1)
        else:
            atom.SetAtomMapNum(0)

    return Chem.MolToSmiles(reactants_mol)


def atom_map_tag_products(mapped_rxn: str) -> str:
    """
    Given atom-mapped reaction, returns disconnection site-tagged product where atoms
    with changed atom environment are represented by [<atom>:1].

    :param mapped_rxn: Atom-mapped reaction SMILES
    :return: SMILES of the product containing tags corresponding to atoms changed in the
        reaction.
    """
    reactants_smiles, _, product_smiles = split_rsmi(mapped_rxn)

    product_mol = Chem.MolFromSmiles(product_smiles)
    atom_list = get_atom_list(reactants_smiles, product_smiles)

    # Set atoms in product with a different environment in reactants to 1
    for atom in product_mol.GetAtoms():
        if atom.GetAtomMapNum() in atom_list:
            atom.SetAtomMapNum(1)
        else:
            atom.SetAtomMapNum(0)

    return Chem.MolToSmiles(product_mol)


def main(args: Optional[Sequence[str]] = None) -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--input")
    parser.add_argument("--in_column", default="RxnSmilesClean")
    parser.add_argument("--out_column", default="products_atom_map_tagged")
    parser.add_argument("--output")

    args = parser.parse_args(args)

    data = pd.read_csv(args.input, sep="\t")

    smiles_col = data[args.in_column].apply(atom_map_tag_products)
    data = data.assign(**{args.out_column: smiles_col})
    data.to_csv(args.output, sep="\t", index=False)


if __name__ == "__main__":
    main()



"""
Module containing script to import ORD dataset to a CSV file
"""

import argparse
import os
import re
from collections import defaultdict
from typing import Optional, Sequence

import pandas as pd
from tqdm import tqdm

try:
    from ord_schema import message_helpers
except ImportError:
    DEPENDENCY_OK = False
else:
    from ord_schema.proto import dataset_pb2

    DEPENDENCY_OK = True


def main(input_args: Optional[Sequence[str]] = None) -> None:
    """Function for command-line tool"""

    if not DEPENDENCY_OK:
        raise ImportError("You need to install the ord-schema package and download the ord-data repository")

    parser = argparse.ArgumentParser("Script to import ORD datasets")
    parser.add_argument("--filenames", nargs="+", help="the files to import")
    parser.add_argument("--output", default="ord_data.csv", help="the output filename")
    parser.add_argument("--include-only", help="name or ID of dataset to extract")
    args = parser.parse_args(input_args)

    if not args.filenames:
        print("No filenames given. Quiting early.")
        return

    include_only = args.include_only
    data = defaultdict(list)
    for filename in tqdm(args.filenames):
        ord_dataset = message_helpers.load_message(filename, dataset_pb2.Dataset)
        if include_only and not (ord_dataset.dataset_id == include_only or ord_dataset.name == include_only):
            continue
        for reaction in tqdm(ord_dataset.reactions, desc=os.path.basename(filename)):
            smiles = message_helpers.get_reaction_smiles(reaction, generate_if_missing=True)
            # Trying to remove spurious spaces in reaction SMILES
            smiles = smiles.strip()
            smiles = re.sub(r"\.\s+", ".", smiles)
            smiles = re.sub(r"\s+\.", ".", smiles)
            # Replace dative bonds
            smiles = smiles.replace("->", "-").replace("<-", "-")
            if smiles.count(">") != 2:
                continue
            data["ID"].append(reaction.reaction_id)
            data["Dataset"].append(ord_dataset.name)
            data["Date"].append(reaction.provenance.experiment_start.value)
            data["ReactionSmiles"].append(smiles)
            try:
                data["Yield"].append(message_helpers.get_product_yield(reaction.outcomes[0].products[0]))
            except IndexError:
                data["Yield"].append(None)

    pd.DataFrame(data).to_csv(args.output, sep="\t", index=False)


if __name__ == "__main__":
    main()



"""Module containing actions that modify the dataframe in some way"""

from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar

import pandas as pd
from rdkit import RDLogger

from rxnutils.pipeline.base import action

rd_logger = RDLogger.logger()
rd_logger.setLevel(RDLogger.CRITICAL)


@action
@dataclass
class DropColumns:
    """Drops columns specified in 'columns'

    yaml example:

    drop_columns:
        columns:
            - NRingChange
            - RingBondMade
    """

    pretty_name: ClassVar[str] = "drop_columns"
    columns: list[str]

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        return data.drop(self.columns, axis=1)

    def __str__(self) -> str:
        return f"{self.pretty_name} (drop columns in specified list, keeps the rest)"


@action
@dataclass
class DropDuplicates:
    """Action for dropping duplicates from dataframe"""

    pretty_name: ClassVar[str] = "drop_duplicates"
    key_columns: list[str]

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        return data.drop_duplicates(subset=self.key_columns)

    def __str__(self) -> str:
        return f"{self.pretty_name} (remove duplicates based on the provided list of columns for comparison)"


@action
@dataclass
class DropRows:
    """Drops rows according to boolean in 'indicator_columns'.
    True => Keep, False => Drop.

    yaml example:

    drop_rows:
        indicator_columns:
            - is_sanitizable
            - is_sanitizable2
    """

    pretty_name: ClassVar[str] = "drop_rows"
    indicator_columns: list[str]

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        for column in self.indicator_columns:
            data = data.iloc[data[column].values]
        return data

    def __str__(self) -> str:
        return f"{self.pretty_name} (drop rows according to booleans specified in 'indicator_columns')"


@action
@dataclass
class KeepColumns:
    """Drops columns not specified in 'columns'

    yaml example:

    keep_columns:
        columns:
            - id
            - classification
            - rsmi_processed
    """

    pretty_name: ClassVar[str] = "keep_columns"
    columns: list[str]

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        drop_columns = list(data.columns)
        for name in self.columns:
            drop_columns.remove(name)

        return data.drop(drop_columns, axis=1)

    def __str__(self) -> str:
        return f"{self.pretty_name} (keeps columns in specified list, drops the rest)"


@action
@dataclass
class RenameColumns:
    """Renames columns specified in 'in_columns' to the names specified in 'out_columns'

    yaml example:

    rename_columns:
        in_columns:
            - column1
            - column2
        out_columns:
            - column1_renamed
            - column2_renamed
    """

    pretty_name: ClassVar[str] = "rename_columns"
    in_columns: list[str]
    out_columns: list[str]

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        rename_mapper = dict(zip(self.in_columns, self.out_columns))
        return data.rename(columns=rename_mapper)

    def __str__(self) -> str:
        return f"{self.pretty_name} (renames the specified columns)"


@action
@dataclass
class QueryDataframe:
    """Uses dataframe query to produce a new (smaller) dataframe. Query must conform to pandas.query()

    yaml file example: Keeping only rows where the has_stereo columns is True:

    query_dataframe:
        query: has_stereo == True
    """

    pretty_name: ClassVar[str] = "query_dataframe"
    query: str

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        return data.query(self.query)

    def __str__(self) -> str:
        return f"{self.pretty_name} (query the dataframe using {self.query} and returns the filtered dataframe)"


@action
@dataclass
class StackColumns:
    """Stacks the specified in_columns under a new column name (out_column),
    multiplies the rest of the columns as appropriate

    yaml control file example:

    stack_columns:
        in_columns:
            - rsmi_processed
            - rsmi_inverted_stereo
        out_column: rsmi_processed
    """

    pretty_name: ClassVar[str] = "stack_columns"
    in_columns: list[str]
    out_column: str = "StackedColumns"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        var_column_name = "D5B447A5B810"  # Temporary name with low risk of collision

        keep_columns = list(data.columns)
        for name in self.in_columns:

            keep_columns.remove(name)

        return data.melt(
            id_vars=keep_columns,
            value_vars=self.in_columns,
            var_name=var_column_name,
            value_name=self.out_column,
        ).drop(var_column_name, axis=1)

    def __str__(self) -> str:
        return f"{self.pretty_name} (stacks two columns into a single column)"


@action
@dataclass
class StackMultiColumns:
    """Stacks the specified target_columns on top of the stack_columns after renaming.

    Example Yaml:

    stack_multi_columns:
        stack_columns:
            - rsmi_inverted_stereo
            - PseudoHash_inverted_stereo
        target_columns:
            - rsmi_processed
            - PseudoHash
    """

    pretty_name: ClassVar[str] = "stack_multi_columns"
    stack_columns: list[str]
    target_columns: list[str]

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        # Wonder if this creates two copies in memory giving memory issues, or reference the original data?
        df_target = data.drop(columns=self.stack_columns)
        df_stack = data.drop(columns=self.target_columns)

        rename_dict = dict(zip(self.stack_columns, self.target_columns))
        df_stack = df_stack.rename(columns=rename_dict)

        return pd.concat([df_target, df_stack])

    def __str__(self) -> str:
        return f"{self.pretty_name} (stacks multiple columns onto multiple columns)"
